# Copyright (C) 2018 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests importing the fire module."""

import sys

import afire
from afire import testutils
import mock


class FireImportTest(testutils.BaseTestCase):
    """Tests importing Fire."""

    def testFire(self):
        with mock.patch.object(sys, "argv", ["commandname"]):
            afire.Fire()

    def testFireMethods(self):
        self.assertIsNotNone(afire.Fire)

    def testNoPrivateMethods(self):
        self.assertTrue(hasattr(afire, "Fire"))
        self.assertFalse(hasattr(afire, "_Fire"))


if __name__ == "__main__":
    testutils.main()



# Copyright (C) 2018 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for formatting.py."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from afire import formatting
from afire import testutils

LINE_LENGTH = 80


class FormattingTest(testutils.BaseTestCase):
    def test_bold(self):
        text = formatting.Bold("hello")
        self.assertIn(text, ["hello", "\x1b[1mhello\x1b[0m"])

    def test_underline(self):
        text = formatting.Underline("hello")
        self.assertIn(text, ["hello", "\x1b[4mhello\x1b[0m"])

    def test_indent(self):
        text = formatting.Indent("hello", spaces=2)
        self.assertEqual("  hello", text)

    def test_indent_multiple_lines(self):
        text = formatting.Indent("hello\nworld", spaces=2)
        self.assertEqual("  hello\n  world", text)

    def test_wrap_one_item(self):
        lines = formatting.WrappedJoin(["rice"])
        self.assertEqual(["rice"], lines)

    def test_wrap_multiple_items(self):
        lines = formatting.WrappedJoin(["rice", "beans", "chicken", "cheese"], width=15)
        self.assertEqual(["rice | beans |", "chicken |", "cheese"], lines)

    def test_ellipsis_truncate(self):
        text = "This is a string"
        truncated_text = formatting.EllipsisTruncate(text=text, available_space=10, line_length=LINE_LENGTH)
        self.assertEqual("This is...", truncated_text)

    def test_ellipsis_truncate_not_enough_space(self):
        text = "This is a string"
        truncated_text = formatting.EllipsisTruncate(text=text, available_space=2, line_length=LINE_LENGTH)
        self.assertEqual("This is a string", truncated_text)

    def test_ellipsis_middle_truncate(self):
        text = "1000000000L"
        truncated_text = formatting.EllipsisMiddleTruncate(text=text, available_space=7, line_length=LINE_LENGTH)
        self.assertEqual("10...0L", truncated_text)

    def test_ellipsis_middle_truncate_not_enough_space(self):
        text = "1000000000L"
        truncated_text = formatting.EllipsisMiddleTruncate(text=text, available_space=2, line_length=LINE_LENGTH)
        self.assertEqual("1000000000L", truncated_text)


if __name__ == "__main__":
    testutils.main()



# Copyright (C) 2018 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module is used for enabling formatting on Windows."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import ctypes
import os
import platform
import subprocess
import sys

try:
    import colorama  # pylint: disable=g-import-not-at-top,  # pytype: disable=import-error

    HAS_COLORAMA = True
except ImportError:
    HAS_COLORAMA = False


def initialize_or_disable():
    """Enables ANSI processing on Windows or disables it as needed."""
    if HAS_COLORAMA:
        wrap = True
        if sys.stdout.isatty() and platform.release() == "10":
            # Enables native ANSI sequences in console.
            # Windows 10, 2016, and 2019 only.

            wrap = False
            kernel32 = ctypes.windll.kernel32  # pytype: disable=module-attr
            enable_virtual_terminal_processing = 0x04
            out_handle = kernel32.GetStdHandle(
                subprocess.STD_OUTPUT_HANDLE
            )  # pylint: disable=line-too-long,  # pytype: disable=module-attr
            # GetConsoleMode fails if the terminal isn't native.
            mode = ctypes.wintypes.DWORD()
            if kernel32.GetConsoleMode(out_handle, ctypes.byref(mode)) == 0:
                wrap = True
            if not mode.value & enable_virtual_terminal_processing:
                if kernel32.SetConsoleMode(out_handle, mode.value | enable_virtual_terminal_processing) == 0:
                    # kernel32.SetConsoleMode to enable ANSI sequences failed
                    wrap = True
        colorama.init(wrap=wrap)
    else:
        os.environ["ANSI_COLORS_DISABLED"] = "1"


if sys.platform.startswith("win"):
    initialize_or_disable()



# Copyright (C) 2018 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the inspectutils module."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import unittest

from afire import inspectutils
from afire import test_components as tc
from afire import testutils

import six


class InspectUtilsTest(testutils.BaseTestCase):
    def testGetFullArgSpec(self):
        spec = inspectutils.GetFullArgSpec(tc.identity)
        self.assertEqual(spec.args, ["arg1", "arg2", "arg3", "arg4"])
        self.assertEqual(spec.defaults, (10, 20))
        self.assertEqual(spec.varargs, "arg5")
        self.assertEqual(spec.varkw, "arg6")
        self.assertEqual(spec.kwonlyargs, [])
        self.assertEqual(spec.kwonlydefaults, {})
        self.assertEqual(spec.annotations, {"arg2": int, "arg4": int})

    @unittest.skipIf(six.PY2, "No keyword arguments in python 2")
    def testGetFullArgSpecPy3(self):
        spec = inspectutils.GetFullArgSpec(tc.py3.identity)
        self.assertEqual(spec.args, ["arg1", "arg2", "arg3", "arg4"])
        self.assertEqual(spec.defaults, (10, 20))
        self.assertEqual(spec.varargs, "arg5")
        self.assertEqual(spec.varkw, "arg10")
        self.assertEqual(spec.kwonlyargs, ["arg6", "arg7", "arg8", "arg9"])
        self.assertEqual(spec.kwonlydefaults, {"arg8": 30, "arg9": 40})
        self.assertEqual(spec.annotations, {"arg2": int, "arg4": int, "arg7": int, "arg9": int})

    def testGetFullArgSpecFromBuiltin(self):
        spec = inspectutils.GetFullArgSpec("test".upper)
        self.assertEqual(spec.args, [])
        self.assertEqual(spec.defaults, ())
        self.assertEqual(spec.kwonlyargs, [])
        self.assertEqual(spec.kwonlydefaults, {})
        self.assertEqual(spec.annotations, {})

    def testGetFullArgSpecFromSlotWrapper(self):
        spec = inspectutils.GetFullArgSpec(tc.NoDefaults)
        self.assertEqual(spec.args, [])
        self.assertEqual(spec.defaults, ())
        self.assertEqual(spec.varargs, None)
        self.assertEqual(spec.varkw, None)
        self.assertEqual(spec.kwonlyargs, [])
        self.assertEqual(spec.kwonlydefaults, {})
        self.assertEqual(spec.annotations, {})

    def testGetFullArgSpecFromNamedTuple(self):
        spec = inspectutils.GetFullArgSpec(tc.NamedTuplePoint)
        self.assertEqual(spec.args, ["x", "y"])
        self.assertEqual(spec.defaults, ())
        self.assertEqual(spec.varargs, None)
        self.assertEqual(spec.varkw, None)
        self.assertEqual(spec.kwonlyargs, [])
        self.assertEqual(spec.kwonlydefaults, {})
        self.assertEqual(spec.annotations, {})

    def testGetFullArgSpecFromNamedTupleSubclass(self):
        spec = inspectutils.GetFullArgSpec(tc.SubPoint)
        self.assertEqual(spec.args, ["x", "y"])
        self.assertEqual(spec.defaults, ())
        self.assertEqual(spec.varargs, None)
        self.assertEqual(spec.varkw, None)
        self.assertEqual(spec.kwonlyargs, [])
        self.assertEqual(spec.kwonlydefaults, {})
        self.assertEqual(spec.annotations, {})

    def testGetFullArgSpecFromClassNoInit(self):
        spec = inspectutils.GetFullArgSpec(tc.OldStyleEmpty)
        self.assertEqual(spec.args, [])
        self.assertEqual(spec.defaults, ())
        self.assertEqual(spec.varargs, None)
        self.assertEqual(spec.varkw, None)
        self.assertEqual(spec.kwonlyargs, [])
        self.assertEqual(spec.kwonlydefaults, {})
        self.assertEqual(spec.annotations, {})

    def testGetFullArgSpecFromMethod(self):
        spec = inspectutils.GetFullArgSpec(tc.NoDefaults().double)
        self.assertEqual(spec.args, ["count"])
        self.assertEqual(spec.defaults, ())
        self.assertEqual(spec.varargs, None)
        self.assertEqual(spec.varkw, None)
        self.assertEqual(spec.kwonlyargs, [])
        self.assertEqual(spec.kwonlydefaults, {})
        self.assertEqual(spec.annotations, {})

    def testInfoOne(self):
        info = inspectutils.Info(1)
        self.assertEqual(info.get("type_name"), "int")
        self.assertEqual(info.get("file"), None)
        self.assertEqual(info.get("line"), None)
        self.assertEqual(info.get("string_form"), "1")

    def testInfoClass(self):
        info = inspectutils.Info(tc.NoDefaults)
        self.assertEqual(info.get("type_name"), "type")
        self.assertIn(os.path.join("fire", "test_components.py"), info.get("file"))
        self.assertGreater(info.get("line"), 0)

    def testInfoClassNoInit(self):
        info = inspectutils.Info(tc.OldStyleEmpty)
        if six.PY2:
            self.assertEqual(info.get("type_name"), "classobj")
        else:
            self.assertEqual(info.get("type_name"), "type")
        self.assertIn(os.path.join("fire", "test_components.py"), info.get("file"))
        self.assertGreater(info.get("line"), 0)

    def testInfoNoDocstring(self):
        info = inspectutils.Info(tc.NoDefaults)
        self.assertEqual(info["docstring"], None, "Docstring should be None")


if __name__ == "__main__":
    testutils.main()



# Copyright (C) 2018 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for fire docstrings module."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from afire import docstrings
from afire import testutils

# pylint: disable=invalid-name
DocstringInfo = docstrings.DocstringInfo
ArgInfo = docstrings.ArgInfo
KwargInfo = docstrings.KwargInfo
# pylint: enable=invalid-name


class DocstringsTest(testutils.BaseTestCase):
    def test_one_line_simple(self):
        docstring = """A simple one line docstring."""
        docstring_info = docstrings.parse(docstring)
        expected_docstring_info = DocstringInfo(
            summary="A simple one line docstring.",
        )
        self.assertEqual(expected_docstring_info, docstring_info)

    def test_one_line_simple_whitespace(self):
        docstring = """
      A simple one line docstring.
    """
        docstring_info = docstrings.parse(docstring)
        expected_docstring_info = DocstringInfo(
            summary="A simple one line docstring.",
        )
        self.assertEqual(expected_docstring_info, docstring_info)

    def test_one_line_too_long(self):
        # pylint: disable=line-too-long
        docstring = """A one line docstring that is both a little too verbose and a little too long so it keeps going well beyond a reasonable length for a one-liner.
    """
        # pylint: enable=line-too-long
        docstring_info = docstrings.parse(docstring)
        expected_docstring_info = DocstringInfo(
            summary="A one line docstring that is both a little too verbose and "
            "a little too long so it keeps going well beyond a reasonable length "
            "for a one-liner.",
        )
        self.assertEqual(expected_docstring_info, docstring_info)

    def test_one_line_runs_over(self):
        # pylint: disable=line-too-long
        docstring = """A one line docstring that is both a little too verbose and a little too long
    so it runs onto a second line.
    """
        # pylint: enable=line-too-long
        docstring_info = docstrings.parse(docstring)
        expected_docstring_info = DocstringInfo(
            summary="A one line docstring that is both a little too verbose and "
            "a little too long so it runs onto a second line.",
        )
        self.assertEqual(expected_docstring_info, docstring_info)

    def test_one_line_runs_over_whitespace(self):
        docstring = """
      A one line docstring that is both a little too verbose and a little too long
      so it runs onto a second line.
    """
        docstring_info = docstrings.parse(docstring)
        expected_docstring_info = DocstringInfo(
            summary="A one line docstring that is both a little too verbose and "
            "a little too long so it runs onto a second line.",
        )
        self.assertEqual(expected_docstring_info, docstring_info)

    def test_google_format_args_only(self):
        docstring = """One line description.

    Args:
      arg1: arg1_description
      arg2: arg2_description
    """
        docstring_info = docstrings.parse(docstring)
        expected_docstring_info = DocstringInfo(
            summary="One line description.",
            args=[
                ArgInfo(name="arg1", description="arg1_description"),
                ArgInfo(name="arg2", description="arg2_description"),
            ],
        )
        self.assertEqual(expected_docstring_info, docstring_info)

    def test_google_format_arg_named_args(self):
        docstring = """
    Args:
      args: arg_description
    """
        docstring_info = docstrings.parse(docstring)
        expected_docstring_info = DocstringInfo(
            args=[
                ArgInfo(name="args", description="arg_description"),
            ]
        )
        self.assertEqual(expected_docstring_info, docstring_info)

    def test_google_format_typed_args_and_returns(self):
        docstring = """Docstring summary.

    This is a longer description of the docstring. It spans multiple lines, as
    is allowed.

    Args:
        param1 (int): The first parameter.
        param2 (str): The second parameter.

    Returns:
        bool: The return value. True for success, False otherwise.
    """
        docstring_info = docstrings.parse(docstring)
        expected_docstring_info = DocstringInfo(
            summary="Docstring summary.",
            description="This is a longer description of the docstring. It spans " "multiple lines, as\nis allowed.",
            args=[
                ArgInfo(name="param1", type="int", description="The first parameter."),
                ArgInfo(name="param2", type="str", description="The second parameter."),
            ],
            returns="bool: The return value. True for success, False otherwise.",
        )
        self.assertEqual(expected_docstring_info, docstring_info)

    def test_google_format_multiline_arg_description(self):
        docstring = """Docstring summary.

    This is a longer description of the docstring. It spans multiple lines, as
    is allowed.

    Args:
        param1 (int): The first parameter.
        param2 (str): The second parameter. This has a lot of text, enough to
        cover two lines.
    """
        docstring_info = docstrings.parse(docstring)
        expected_docstring_info = DocstringInfo(
            summary="Docstring summary.",
            description="This is a longer description of the docstring. It spans " "multiple lines, as\nis allowed.",
            args=[
                ArgInfo(name="param1", type="int", description="The first parameter."),
                ArgInfo(
                    name="param2",
                    type="str",
                    description="The second parameter. This has a lot of text, " "enough to cover two lines.",
                ),
            ],
        )
        self.assertEqual(expected_docstring_info, docstring_info)

    def test_rst_format_typed_args_and_returns(self):
        docstring = """Docstring summary.

    This is a longer description of the docstring. It spans across multiple
    lines.

    :param arg1: Description of arg1.
    :type arg1: str.
    :param arg2: Description of arg2.
    :type arg2: bool.
    :returns:  int -- description of the return value.
    :raises: AttributeError, KeyError
    """
        docstring_info = docstrings.parse(docstring)
        expected_docstring_info = DocstringInfo(
            summary="Docstring summary.",
            description="This is a longer description of the docstring. It spans " "across multiple\nlines.",
            args=[
                ArgInfo(name="arg1", type="str", description="Description of arg1."),
                ArgInfo(name="arg2", type="bool", description="Description of arg2."),
            ],
            returns="int -- description of the return value.",
            raises="AttributeError, KeyError",
        )
        self.assertEqual(expected_docstring_info, docstring_info)

    def test_numpy_format_typed_args_and_returns(self):
        docstring = """Docstring summary.

    This is a longer description of the docstring. It spans across multiple
    lines.

    Parameters
    ----------
    param1 : int
        The first parameter.
    param2 : str
        The second parameter.

    Returns
    -------
    bool
        True if successful, False otherwise.
    """
        docstring_info = docstrings.parse(docstring)
        expected_docstring_info = DocstringInfo(
            summary="Docstring summary.",
            description="This is a longer description of the docstring. It spans " "across multiple\nlines.",
            args=[
                ArgInfo(name="param1", type="int", description="The first parameter."),
                ArgInfo(name="param2", type="str", description="The second parameter."),
            ],
            # TODO(dbieber): Support return type.
            returns="bool True if successful, False otherwise.",
        )
        self.assertEqual(expected_docstring_info, docstring_info)

    def test_numpy_format_multiline_arg_description(self):
        docstring = """Docstring summary.

    This is a longer description of the docstring. It spans across multiple
    lines.

    Parameters
    ----------
    param1 : int
        The first parameter.
    param2 : str
        The second parameter. This has a lot of text, enough to cover two
        lines.
    """
        docstring_info = docstrings.parse(docstring)
        expected_docstring_info = DocstringInfo(
            summary="Docstring summary.",
            description="This is a longer description of the docstring. It spans " "across multiple\nlines.",
            args=[
                ArgInfo(name="param1", type="int", description="The first parameter."),
                ArgInfo(
                    name="param2",
                    type="str",
                    description="The second parameter. This has a lot of text, " "enough to cover two lines.",
                ),
            ],
        )
        self.assertEqual(expected_docstring_info, docstring_info)

    def test_multisection_docstring(self):
        docstring = """Docstring summary.

    This is the first section of a docstring description.

    This is the second section of a docstring description. This docstring
    description has just two sections.
    """
        docstring_info = docstrings.parse(docstring)
        expected_docstring_info = DocstringInfo(
            summary="Docstring summary.",
            description="This is the first section of a docstring description."
            "\n\n"
            "This is the second section of a docstring description. This docstring"
            "\n"
            "description has just two sections.",
        )
        self.assertEqual(expected_docstring_info, docstring_info)

    def test_google_section_with_blank_first_line(self):
        docstring = """Inspired by requests HTTPAdapter docstring.

    :param x: Simple param.

    Usage:

      >>> import requests
    """
        docstring_info = docstrings.parse(docstring)
        self.assertEqual("Inspired by requests HTTPAdapter docstring.", docstring_info.summary)

    def test_ill_formed_docstring(self):
        docstring = """Docstring summary.

    args: raises ::
    :
    pathological docstrings should not fail, and ideally should behave
    reasonably.
    """
        docstrings.parse(docstring)

    def test_strip_blank_lines(self):
        lines = ["   ", "  foo  ", "   "]
        expected_output = ["  foo  "]

        self.assertEqual(expected_output, docstrings._strip_blank_lines(lines))  # pylint: disable=protected-access

    def test_numpy_colon_in_description(self):
        docstring = """
     Greets name.

     Arguments
     ---------
     name : str
         name, default : World
     arg2 : int
         arg2, default:None
     arg3 : bool
     """
        docstring_info = docstrings.parse(docstring)
        expected_docstring_info = DocstringInfo(
            summary="Greets name.",
            description=None,
            args=[
                ArgInfo(name="name", type="str", description="name, default : World"),
                ArgInfo(name="arg2", type="int", description="arg2, default:None"),
                ArgInfo(name="arg3", type="bool", description=None),
            ],
        )
        self.assertEqual(expected_docstring_info, docstring_info)

    def test_rst_format_typed_args_and_kwargs(self):
        docstring = """Docstring summary.

    :param arg1: Description of arg1.
    :type arg1: str.
    :key arg2: Description of arg2.
    :type arg2: bool.
    :key arg3: Description of arg3.
    :type arg3: str.
    """
        docstring_info = docstrings.parse(docstring)
        expected_docstring_info = DocstringInfo(
            summary="Docstring summary.",
            args=[
                ArgInfo(name="arg1", type="str", description="Description of arg1."),
                KwargInfo(name="arg2", type="bool", description="Description of arg2."),
                KwargInfo(name="arg3", type="str", description="Description of arg3."),
            ],
        )
        self.assertEqual(expected_docstring_info, docstring_info)


if __name__ == "__main__":
    testutils.main()



# Copyright (C) 2018 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the fire module."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import sys

import afire
from afire import test_components as tc
from afire import testutils

import mock
import six


class FireTest(testutils.BaseTestCase):
    def testFire(self):
        with mock.patch.object(sys, "argv", ["progname"]):
            afire.Fire(tc.Empty)
            afire.Fire(tc.OldStyleEmpty)
            afire.Fire(tc.WithInit)
        # Test both passing command as a sequence and as a string.
        self.assertEqual(afire.Fire(tc.NoDefaults, command="triple 4"), 12)
        self.assertEqual(afire.Fire(tc.WithDefaults, command=("double", "2")), 4)
        self.assertEqual(afire.Fire(tc.WithDefaults, command=["triple", "4"]), 12)
        self.assertEqual(afire.Fire(tc.OldStyleWithDefaults, command=["double", "2"]), 4)
        self.assertEqual(afire.Fire(tc.OldStyleWithDefaults, command=["triple", "4"]), 12)

    def testFirePositionalCommand(self):
        # Test passing command as a positional argument.
        self.assertEqual(afire.Fire(tc.NoDefaults, "double 2"), 4)
        self.assertEqual(afire.Fire(tc.NoDefaults, ["double", "2"]), 4)

    def testFireInvalidCommandArg(self):
        with self.assertRaises(ValueError):
            # This is not a valid command.
            afire.Fire(tc.WithDefaults, command=10)

    def testFireDefaultName(self):
        with mock.patch.object(sys, "argv", [os.path.join("python-fire", "fire", "base_filename.py")]):
            with self.assertOutputMatches(stdout="SYNOPSIS.*base_filename.py", stderr=None):
                afire.Fire(tc.Empty)

    def testFireNoArgs(self):
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["ten"]), 10)

    def testFireExceptions(self):
        # Exceptions of Fire are printed to stderr and a FireExit is raised.
        with self.assertRaisesFireExit(2):
            afire.Fire(tc.Empty, command=["nomethod"])  # Member doesn't exist.
        with self.assertRaisesFireExit(2):
            afire.Fire(tc.NoDefaults, command=["double"])  # Missing argument.
        with self.assertRaisesFireExit(2):
            afire.Fire(tc.TypedProperties, command=["delta", "x"])  # Missing key.

        # Exceptions of the target components are still raised.
        with self.assertRaises(ZeroDivisionError):
            afire.Fire(tc.NumberDefaults, command=["reciprocal", "0.0"])

    def testFireNamedArgs(self):
        self.assertEqual(afire.Fire(tc.WithDefaults, command=["double", "--count", "5"]), 10)
        self.assertEqual(afire.Fire(tc.WithDefaults, command=["triple", "--count", "5"]), 15)
        self.assertEqual(afire.Fire(tc.OldStyleWithDefaults, command=["double", "--count", "5"]), 10)
        self.assertEqual(afire.Fire(tc.OldStyleWithDefaults, command=["triple", "--count", "5"]), 15)

    def testFireNamedArgsSingleHyphen(self):
        self.assertEqual(afire.Fire(tc.WithDefaults, command=["double", "-count", "5"]), 10)
        self.assertEqual(afire.Fire(tc.WithDefaults, command=["triple", "-count", "5"]), 15)
        self.assertEqual(afire.Fire(tc.OldStyleWithDefaults, command=["double", "-count", "5"]), 10)
        self.assertEqual(afire.Fire(tc.OldStyleWithDefaults, command=["triple", "-count", "5"]), 15)

    def testFireNamedArgsWithEquals(self):
        self.assertEqual(afire.Fire(tc.WithDefaults, command=["double", "--count=5"]), 10)
        self.assertEqual(afire.Fire(tc.WithDefaults, command=["triple", "--count=5"]), 15)

    def testFireNamedArgsWithEqualsSingleHyphen(self):
        self.assertEqual(afire.Fire(tc.WithDefaults, command=["double", "-count=5"]), 10)
        self.assertEqual(afire.Fire(tc.WithDefaults, command=["triple", "-count=5"]), 15)

    def testFireAllNamedArgs(self):
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["sum", "1", "2"]), 5)
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["sum", "--alpha", "1", "2"]), 5)
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["sum", "--beta", "1", "2"]), 4)
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["sum", "1", "--alpha", "2"]), 4)
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["sum", "1", "--beta", "2"]), 5)
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["sum", "--alpha", "1", "--beta", "2"]), 5)
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["sum", "--beta", "1", "--alpha", "2"]), 4)

    def testFireAllNamedArgsOneMissing(self):
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["sum"]), 0)
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["sum", "1"]), 1)
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["sum", "--alpha", "1"]), 1)
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["sum", "--beta", "2"]), 4)

    def testFirePartialNamedArgs(self):
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "1", "2"]), (1, 2))
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "--alpha", "1", "2"]), (1, 2))
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "--beta", "1", "2"]), (2, 1))
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "1", "--alpha", "2"]), (2, 1))
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "1", "--beta", "2"]), (1, 2))
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "--alpha", "1", "--beta", "2"]), (1, 2))
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "--beta", "1", "--alpha", "2"]), (2, 1))

    def testFirePartialNamedArgsOneMissing(self):
        # Errors are written to standard out and a FireExit is raised.
        with self.assertRaisesFireExit(2):
            afire.Fire(tc.MixedDefaults, command=["identity"])  # Identity needs an arg.

        with self.assertRaisesFireExit(2):
            # Identity needs a value for alpha.
            afire.Fire(tc.MixedDefaults, command=["identity", "--beta", "2"])

        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "1"]), (1, "0"))
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "--alpha", "1"]), (1, "0"))

    def testFireAnnotatedArgs(self):
        self.assertEqual(afire.Fire(tc.Annotations, command=["double", "5"]), 10)
        self.assertEqual(afire.Fire(tc.Annotations, command=["triple", "5"]), 15)

    @testutils.skipIf(six.PY2, "Keyword-only arguments not in Python 2.")
    def testFireKeywordOnlyArgs(self):
        with self.assertRaisesFireExit(2):
            # Keyword arguments must be passed with flag syntax.
            afire.Fire(tc.py3.KeywordOnly, command=["double", "5"])

        self.assertEqual(afire.Fire(tc.py3.KeywordOnly, command=["double", "--count", "5"]), 10)
        self.assertEqual(afire.Fire(tc.py3.KeywordOnly, command=["triple", "--count", "5"]), 15)

    def testFireProperties(self):
        self.assertEqual(afire.Fire(tc.TypedProperties, command=["alpha"]), True)
        self.assertEqual(afire.Fire(tc.TypedProperties, command=["beta"]), (1, 2, 3))

    def testFireRecursion(self):
        self.assertEqual(afire.Fire(tc.TypedProperties, command=["charlie", "double", "hello"]), "hellohello")
        self.assertEqual(afire.Fire(tc.TypedProperties, command=["charlie", "triple", "w"]), "www")

    def testFireVarArgs(self):
        self.assertEqual(afire.Fire(tc.VarArgs, command=["cumsums", "a", "b", "c", "d"]), ["a", "ab", "abc", "abcd"])
        self.assertEqual(afire.Fire(tc.VarArgs, command=["cumsums", "1", "2", "3", "4"]), [1, 3, 6, 10])

    def testFireVarArgsWithNamedArgs(self):
        self.assertEqual(afire.Fire(tc.VarArgs, command=["varchars", "1", "2", "c", "d"]), (1, 2, "cd"))
        self.assertEqual(afire.Fire(tc.VarArgs, command=["varchars", "3", "4", "c", "d", "e"]), (3, 4, "cde"))

    def testFireKeywordArgs(self):
        self.assertEqual(
            afire.Fire(tc.Kwargs, command=["props", "--name", "David", "--age", "24"]), {"name": "David", "age": 24}
        )
        # Run this test both with a list command and a string command.
        self.assertEqual(
            afire.Fire(
                tc.Kwargs, command=["props", "--message", '"This is a message it has -- in it"']
            ),  # Quotes stripped
            {"message": "This is a message it has -- in it"},
        )
        self.assertEqual(
            afire.Fire(tc.Kwargs, command=["props", "--message", "This is a message it has -- in it"]),
            {"message": "This is a message it has -- in it"},
        )
        self.assertEqual(
            afire.Fire(tc.Kwargs, command='props --message "This is a message it has -- in it"'),
            {"message": "This is a message it has -- in it"},
        )
        self.assertEqual(afire.Fire(tc.Kwargs, command=["upper", "--alpha", "A", "--beta", "B"]), "ALPHA BETA")
        self.assertEqual(
            afire.Fire(tc.Kwargs, command=["upper", "--alpha", "A", "--beta", "B", "-", "lower"]), "alpha beta"
        )

    def testFireKeywordArgsWithMissingPositionalArgs(self):
        self.assertEqual(
            afire.Fire(tc.Kwargs, command=["run", "Hello", "World", "--cell", "is"]), ("Hello", "World", {"cell": "is"})
        )
        self.assertEqual(
            afire.Fire(tc.Kwargs, command=["run", "Hello", "--cell", "ok"]), ("Hello", None, {"cell": "ok"})
        )

    def testFireObject(self):
        self.assertEqual(afire.Fire(tc.WithDefaults(), command=["double", "--count", "5"]), 10)
        self.assertEqual(afire.Fire(tc.WithDefaults(), command=["triple", "--count", "5"]), 15)

    def testFireDict(self):
        component = {
            "double": lambda x=0: 2 * x,
            "cheese": "swiss",
        }
        self.assertEqual(afire.Fire(component, command=["double", "5"]), 10)
        self.assertEqual(afire.Fire(component, command=["cheese"]), "swiss")

    def testFireObjectWithDict(self):
        self.assertEqual(afire.Fire(tc.TypedProperties, command=["delta", "echo"]), "E")
        self.assertEqual(afire.Fire(tc.TypedProperties, command=["delta", "echo", "lower"]), "e")
        self.assertIsInstance(afire.Fire(tc.TypedProperties, command=["delta", "nest"]), dict)
        self.assertEqual(afire.Fire(tc.TypedProperties, command=["delta", "nest", "0"]), "a")

    def testFireSet(self):
        component = tc.simple_set()
        result = afire.Fire(component, command=[])
        self.assertEqual(len(result), 3)

    def testFireFrozenset(self):
        component = tc.simple_frozenset()
        result = afire.Fire(component, command=[])
        self.assertEqual(len(result), 3)

    def testFireList(self):
        component = ["zero", "one", "two", "three"]
        self.assertEqual(afire.Fire(component, command=["2"]), "two")
        self.assertEqual(afire.Fire(component, command=["3"]), "three")
        self.assertEqual(afire.Fire(component, command=["-1"]), "three")

    def testFireObjectWithList(self):
        self.assertEqual(afire.Fire(tc.TypedProperties, command=["echo", "0"]), "alex")
        self.assertEqual(afire.Fire(tc.TypedProperties, command=["echo", "1"]), "bethany")

    def testFireObjectWithTuple(self):
        self.assertEqual(afire.Fire(tc.TypedProperties, command=["fox", "0"]), "carry")
        self.assertEqual(afire.Fire(tc.TypedProperties, command=["fox", "1"]), "divide")

    def testFireObjectWithListAsObject(self):
        self.assertEqual(afire.Fire(tc.TypedProperties, command=["echo", "count", "bethany"]), 1)

    def testFireObjectWithTupleAsObject(self):
        self.assertEqual(afire.Fire(tc.TypedProperties, command=["fox", "count", "divide"]), 1)

    def testFireNoComponent(self):
        self.assertEqual(afire.Fire(command=["tc", "WithDefaults", "double", "10"]), 20)
        last_char = lambda text: text[-1]  # pylint: disable=unused-variable
        self.assertEqual(afire.Fire(command=["last_char", '"Hello"']), "o")
        self.assertEqual(afire.Fire(command=["last-char", '"World"']), "d")
        rset = lambda count=0: set(range(count))  # pylint: disable=unused-variable
        self.assertEqual(afire.Fire(command=["rset", "5"]), {0, 1, 2, 3, 4})

    def testFireUnderscores(self):
        self.assertEqual(afire.Fire(tc.Underscores, command=["underscore-example"]), "fish fingers")
        self.assertEqual(afire.Fire(tc.Underscores, command=["underscore_example"]), "fish fingers")

    def testFireUnderscoresInArg(self):
        self.assertEqual(afire.Fire(tc.Underscores, command=["underscore-function", "example"]), "example")
        self.assertEqual(afire.Fire(tc.Underscores, command=["underscore_function", "--underscore-arg=score"]), "score")
        self.assertEqual(afire.Fire(tc.Underscores, command=["underscore_function", "--underscore_arg=score"]), "score")

    def testBoolParsing(self):
        self.assertEqual(afire.Fire(tc.BoolConverter, command=["as-bool", "True"]), True)
        self.assertEqual(afire.Fire(tc.BoolConverter, command=["as-bool", "False"]), False)
        self.assertEqual(afire.Fire(tc.BoolConverter, command=["as-bool", "--arg=True"]), True)
        self.assertEqual(afire.Fire(tc.BoolConverter, command=["as-bool", "--arg=False"]), False)
        self.assertEqual(afire.Fire(tc.BoolConverter, command=["as-bool", "--arg"]), True)
        self.assertEqual(afire.Fire(tc.BoolConverter, command=["as-bool", "--noarg"]), False)

    def testBoolParsingContinued(self):
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "True", "False"]), (True, False))
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "--alpha=False", "10"]), (False, 10))
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "--alpha", "--beta", "10"]), (True, 10))
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "--alpha", "--beta=10"]), (True, 10))
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "--noalpha", "--beta"]), (False, True))
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "10", "--beta"]), (10, True))

    def testBoolParsingSingleHyphen(self):
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "-alpha=False", "10"]), (False, 10))
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "-alpha", "-beta", "10"]), (True, 10))
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "-alpha", "-beta=10"]), (True, 10))
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "-noalpha", "-beta"]), (False, True))
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "-alpha", "-10", "-beta"]), (-10, True))

    def testBoolParsingLessExpectedCases(self):
        # Note: Does not return (True, 10).
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "--alpha", "10"]), (10, "0"))
        # To get (True, 10), use one of the following:
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "--alpha", "--beta=10"]), (True, 10))
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "True", "10"]), (True, 10))

        # Note: Does not return (True, '--test') or ('--test', 0).
        with self.assertRaisesFireExit(2):
            afire.Fire(tc.MixedDefaults, command=["identity", "--alpha", "--test"])

        self.assertEqual(
            afire.Fire(tc.MixedDefaults, command=["identity", "--alpha", "True", '"--test"']), (True, "--test")
        )
        # To get ('--test', '0'), use one of the following:
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "--alpha=--test"]), ("--test", "0"))
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=r"identity --alpha \"--test\""), ("--test", "0"))

    def testSingleCharFlagParsing(self):
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "-a"]), (True, "0"))
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "-a", "--beta=10"]), (True, 10))
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "-a", "-b"]), (True, True))
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "-a", "42", "-b"]), (42, True))
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "-a", "42", "-b", "10"]), (42, 10))
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "--alpha", "True", "-b", "10"]), (True, 10))
        with self.assertRaisesFireExit(2):
            # This test attempts to use an ambiguous shortcut flag on a function with
            # a naming conflict for the shortcut, triggering a FireError.
            afire.Fire(tc.SimilarArgNames, command=["identity", "-b"])

    def testSingleCharFlagParsingEqualSign(self):
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "-a=True"]), (True, "0"))
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "-a=3", "--beta=10"]), (3, 10))
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "-a=False", "-b=15"]), (False, 15))
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "-a", "42", "-b=12"]), (42, 12))
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "-a=42", "-b", "10"]), (42, 10))

    def testSingleCharFlagParsingExactMatch(self):
        self.assertEqual(afire.Fire(tc.SimilarArgNames, command=["identity2", "-a"]), (True, None))
        self.assertEqual(afire.Fire(tc.SimilarArgNames, command=["identity2", "-a=10"]), (10, None))
        self.assertEqual(afire.Fire(tc.SimilarArgNames, command=["identity2", "--a"]), (True, None))
        self.assertEqual(afire.Fire(tc.SimilarArgNames, command=["identity2", "-alpha"]), (None, True))
        self.assertEqual(afire.Fire(tc.SimilarArgNames, command=["identity2", "-a", "-alpha"]), (True, True))

    def testSingleCharFlagParsingCapitalLetter(self):
        self.assertEqual(afire.Fire(tc.CapitalizedArgNames, command=["sum", "-D", "5", "-G", "10"]), 15)

    def testBoolParsingWithNo(self):
        # In these examples --nothing always refers to the nothing argument:
        def fn1(thing, nothing):
            return thing, nothing

        self.assertEqual(afire.Fire(fn1, command=["--thing", "--nothing"]), (True, True))
        self.assertEqual(afire.Fire(fn1, command=["--thing", "--nonothing"]), (True, False))

        with self.assertRaisesFireExit(2):
            # In this case nothing=False (since rightmost setting of a flag gets
            # precedence), but it errors because thing has no value.
            afire.Fire(fn1, command=["--nothing", "--nonothing"])

        # In these examples, --nothing sets thing=False:
        def fn2(thing, **kwargs):
            return thing, kwargs

        self.assertEqual(afire.Fire(fn2, command=["--thing"]), (True, {}))
        self.assertEqual(afire.Fire(fn2, command=["--nothing"]), (False, {}))
        with self.assertRaisesFireExit(2):
            # In this case, nothing=True, but it errors because thing has no value.
            afire.Fire(fn2, command=["--nothing=True"])
        self.assertEqual(afire.Fire(fn2, command=["--nothing", "--nothing=True"]), (False, {"nothing": True}))

        def fn3(arg, **kwargs):
            return arg, kwargs

        self.assertEqual(afire.Fire(fn3, command=["--arg=value", "--thing"]), ("value", {"thing": True}))
        self.assertEqual(afire.Fire(fn3, command=["--arg=value", "--nothing"]), ("value", {"thing": False}))
        self.assertEqual(afire.Fire(fn3, command=["--arg=value", "--nonothing"]), ("value", {"nothing": False}))

    def testTraceFlag(self):
        with self.assertRaisesFireExit(0, "Fire trace:\n"):
            afire.Fire(tc.BoolConverter, command=["as-bool", "True", "--", "--trace"])
        with self.assertRaisesFireExit(0, "Fire trace:\n"):
            afire.Fire(tc.BoolConverter, command=["as-bool", "True", "--", "-t"])
        with self.assertRaisesFireExit(0, "Fire trace:\n"):
            afire.Fire(tc.BoolConverter, command=["--", "--trace"])

    def testHelpFlag(self):
        with self.assertRaisesFireExit(0):
            afire.Fire(tc.BoolConverter, command=["as-bool", "True", "--", "--help"])
        with self.assertRaisesFireExit(0):
            afire.Fire(tc.BoolConverter, command=["as-bool", "True", "--", "-h"])
        with self.assertRaisesFireExit(0):
            afire.Fire(tc.BoolConverter, command=["--", "--help"])

    def testHelpFlagAndTraceFlag(self):
        with self.assertRaisesFireExit(0, "Fire trace:\n.*SYNOPSIS"):
            afire.Fire(tc.BoolConverter, command=["as-bool", "True", "--", "--help", "--trace"])
        with self.assertRaisesFireExit(0, "Fire trace:\n.*SYNOPSIS"):
            afire.Fire(tc.BoolConverter, command=["as-bool", "True", "--", "-h", "-t"])
        with self.assertRaisesFireExit(0, "Fire trace:\n.*SYNOPSIS"):
            afire.Fire(tc.BoolConverter, command=["--", "-h", "--trace"])

    def testTabCompletionNoName(self):
        completion_script = afire.Fire(tc.NoDefaults, command=["--", "--completion"])
        self.assertIn("double", completion_script)
        self.assertIn("triple", completion_script)

    def testTabCompletion(self):
        completion_script = afire.Fire(tc.NoDefaults, command=["--", "--completion"], name="c")
        self.assertIn("double", completion_script)
        self.assertIn("triple", completion_script)

    def testTabCompletionWithDict(self):
        actions = {"multiply": lambda a, b: a * b}
        completion_script = afire.Fire(actions, command=["--", "--completion"], name="actCLI")
        self.assertIn("actCLI", completion_script)
        self.assertIn("multiply", completion_script)

    def testBasicSeparator(self):
        # '-' is the default separator.
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "+", "_"]), ("+", "_"))
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["identity", "_", "+", "-"]), ("_", "+"))

        # If we change the separator we can use '-' as an argument.
        self.assertEqual(
            afire.Fire(tc.MixedDefaults, command=["identity", "-", "_", "--", "--separator", "&"]), ("-", "_")
        )

        # The separator triggers a function call, but there aren't enough arguments.
        with self.assertRaisesFireExit(2):
            afire.Fire(tc.MixedDefaults, command=["identity", "-", "_", "+"])

    def testNonComparable(self):
        """Fire should work with classes that disallow comparisons."""
        # Make sure this test passes both with a string command or a list command.
        self.assertIsInstance(afire.Fire(tc.NonComparable, command=""), tc.NonComparable)
        self.assertIsInstance(afire.Fire(tc.NonComparable, command=[]), tc.NonComparable)

        # The first separator instantiates the NonComparable object.
        # The second separator causes Fire to check if the separator was necessary.
        self.assertIsInstance(afire.Fire(tc.NonComparable, command=["-", "-"]), tc.NonComparable)

    def testExtraSeparators(self):
        self.assertEqual(
            afire.Fire(tc.ReturnsObj, command=["get-obj", "arg1", "arg2", "-", "-", "as-bool", "True"]), True
        )
        self.assertEqual(
            afire.Fire(tc.ReturnsObj, command=["get-obj", "arg1", "arg2", "-", "-", "-", "as-bool", "True"]), True
        )

    def testSeparatorForChaining(self):
        # Without a separator all args are consumed by get_obj.
        self.assertIsInstance(
            afire.Fire(tc.ReturnsObj, command=["get-obj", "arg1", "arg2", "as-bool", "True"]), tc.BoolConverter
        )
        # With a separator only the preceding args are consumed by get_obj.
        self.assertEqual(afire.Fire(tc.ReturnsObj, command=["get-obj", "arg1", "arg2", "-", "as-bool", "True"]), True)
        self.assertEqual(
            afire.Fire(
                tc.ReturnsObj, command=["get-obj", "arg1", "arg2", "&", "as-bool", "True", "--", "--separator", "&"]
            ),
            True,
        )
        self.assertEqual(
            afire.Fire(tc.ReturnsObj, command=["get-obj", "arg1", "$$", "as-bool", "True", "--", "--separator", "$$"]),
            True,
        )

    def testNegativeNumbers(self):
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["sum", "--alpha", "-3", "--beta", "-4"]), -11)

    def testFloatForExpectedInt(self):
        self.assertEqual(afire.Fire(tc.MixedDefaults, command=["sum", "--alpha", "2.2", "--beta", "3.0"]), 8.2)
        self.assertEqual(afire.Fire(tc.NumberDefaults, command=["integer_reciprocal", "--divisor", "5.0"]), 0.2)
        self.assertEqual(afire.Fire(tc.NumberDefaults, command=["integer_reciprocal", "4.0"]), 0.25)

    def testClassInstantiation(self):
        self.assertIsInstance(afire.Fire(tc.InstanceVars, command=["--arg1=a1", "--arg2=a2"]), tc.InstanceVars)
        with self.assertRaisesFireExit(2):
            # Cannot instantiate a class with positional args.
            afire.Fire(tc.InstanceVars, command=["a1", "a2"])

    def testTraceErrors(self):
        # Class needs additional value but runs out of args.
        with self.assertRaisesFireExit(2):
            afire.Fire(tc.InstanceVars, command=["a1"])
        with self.assertRaisesFireExit(2):
            afire.Fire(tc.InstanceVars, command=["--arg1=a1"])

        # Routine needs additional value but runs out of args.
        with self.assertRaisesFireExit(2):
            afire.Fire(tc.InstanceVars, command=["a1", "a2", "-", "run", "b1"])
        with self.assertRaisesFireExit(2):
            afire.Fire(tc.InstanceVars, command=["--arg1=a1", "--arg2=a2", "-", "run b1"])

        # Extra args cannot be consumed.
        with self.assertRaisesFireExit(2):
            afire.Fire(tc.InstanceVars, command=["a1", "a2", "-", "run", "b1", "b2", "b3"])
        with self.assertRaisesFireExit(2):
            afire.Fire(tc.InstanceVars, command=["--arg1=a1", "--arg2=a2", "-", "run", "b1", "b2", "b3"])

        # Cannot find member to access.
        with self.assertRaisesFireExit(2):
            afire.Fire(tc.InstanceVars, command=["a1", "a2", "-", "jog"])
        with self.assertRaisesFireExit(2):
            afire.Fire(tc.InstanceVars, command=["--arg1=a1", "--arg2=a2", "-", "jog"])

    def testClassWithDefaultMethod(self):
        self.assertEqual(afire.Fire(tc.DefaultMethod, command=["double", "10"]), 20)

    def testClassWithInvalidProperty(self):
        self.assertEqual(afire.Fire(tc.InvalidProperty, command=["double", "10"]), 20)

    @testutils.skipIf(sys.version_info[0:2] <= (3, 4), "Cannot inspect wrapped signatures in Python 2 or 3.4.")
    def testHelpKwargsDecorator(self):
        # Issue #190, follow the wrapped method instead of crashing.
        with self.assertRaisesFireExit(0):
            afire.Fire(tc.decorated_method, command=["-h"])
        with self.assertRaisesFireExit(0):
            afire.Fire(tc.decorated_method, command=["--help"])

    @testutils.skipIf(six.PY2, "Asyncio not available in Python 2.")
    def testFireAsyncio(self):
        self.assertEqual(afire.Fire(tc.py3.WithAsyncio, command=["double", "--count", "10"]), 20)


if __name__ == "__main__":
    testutils.main()



# -*- coding: utf-8 -*- #
# Copyright 2013 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Some general file utilities used that can be used by the Cloud SDK."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import os

from afire.console import encoding as encoding_util
from afire.console import platforms

import six


def _GetSystemPath():
    """Returns properly encoded system PATH variable string."""
    return encoding_util.GetEncodedValue(os.environ, "PATH")


def _FindExecutableOnPath(executable, path, pathext):
    """Internal function to a find an executable.

    Args:
      executable: The name of the executable to find.
      path: A list of directories to search separated by 'os.pathsep'.
      pathext: An iterable of file name extensions to use.

    Returns:
      str, the path to a file on `path` with name `executable` + `p` for
        `p` in `pathext`.

    Raises:
      ValueError: invalid input.
    """

    if isinstance(pathext, six.string_types):
        raise ValueError(
            "_FindExecutableOnPath(..., pathext='{0}') failed "
            "because pathext must be an iterable of strings, but got "
            "a string.".format(pathext)
        )

    # Prioritize preferred extension over earlier in path.
    for ext in pathext:
        for directory in path.split(os.pathsep):
            # Windows can have paths quoted.
            directory = directory.strip('"')
            full = os.path.normpath(os.path.join(directory, executable) + ext)
            # On Windows os.access(full, os.X_OK) is always True.
            if os.path.isfile(full) and os.access(full, os.X_OK):
                return full
    return None


def _PlatformExecutableExtensions(platform):
    if platform == platforms.OperatingSystem.WINDOWS:
        return (".exe", ".cmd", ".bat", ".com", ".ps1")
    else:
        return ("", ".sh")


def FindExecutableOnPath(executable, path=None, pathext=None, allow_extensions=False):
    """Searches for `executable` in the directories listed in `path` or $PATH.

    Executable must not contain a directory or an extension.

    Args:
      executable: The name of the executable to find.
      path: A list of directories to search separated by 'os.pathsep'.  If None
        then the system PATH is used.
      pathext: An iterable of file name extensions to use.  If None then
        platform specific extensions are used.
      allow_extensions: A boolean flag indicating whether extensions in the
        executable are allowed.

    Returns:
      The path of 'executable' (possibly with a platform-specific extension) if
      found and executable, None if not found.

    Raises:
      ValueError: if executable has a path or an extension, and extensions are
        not allowed, or if there's an internal error.
    """

    if not allow_extensions and os.path.splitext(executable)[1]:
        raise ValueError(
            "FindExecutableOnPath({0},...) failed because first "
            "argument must not have an extension.".format(executable)
        )

    if os.path.dirname(executable):
        raise ValueError(
            "FindExecutableOnPath({0},...) failed because first " "argument must not have a path.".format(executable)
        )

    if path is None:
        effective_path = _GetSystemPath()
    else:
        effective_path = path
    effective_pathext = (
        pathext if pathext is not None else _PlatformExecutableExtensions(platforms.OperatingSystem.Current())
    )

    return _FindExecutableOnPath(executable, effective_path, effective_pathext)



# Copyright (C) 2018 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the parser module."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from afire import parser
from afire import testutils


class ParserTest(testutils.BaseTestCase):
    def testCreateParser(self):
        self.assertIsNotNone(parser.CreateParser())

    def testSeparateFlagArgs(self):
        self.assertEqual(parser.SeparateFlagArgs([]), ([], []))
        self.assertEqual(parser.SeparateFlagArgs(["a", "b"]), (["a", "b"], []))
        self.assertEqual(parser.SeparateFlagArgs(["a", "b", "--"]), (["a", "b"], []))
        self.assertEqual(parser.SeparateFlagArgs(["a", "b", "--", "c"]), (["a", "b"], ["c"]))
        self.assertEqual(parser.SeparateFlagArgs(["--"]), ([], []))
        self.assertEqual(parser.SeparateFlagArgs(["--", "c", "d"]), ([], ["c", "d"]))
        self.assertEqual(parser.SeparateFlagArgs(["a", "b", "--", "c", "d"]), (["a", "b"], ["c", "d"]))
        self.assertEqual(parser.SeparateFlagArgs(["a", "b", "--", "c", "d", "--"]), (["a", "b", "--", "c", "d"], []))
        self.assertEqual(parser.SeparateFlagArgs(["a", "b", "--", "c", "--", "d"]), (["a", "b", "--", "c"], ["d"]))

    def testDefaultParseValueStrings(self):
        self.assertEqual(parser.DefaultParseValue("hello"), "hello")
        self.assertEqual(parser.DefaultParseValue("path/file.jpg"), "path/file.jpg")
        self.assertEqual(parser.DefaultParseValue("hello world"), "hello world")
        self.assertEqual(parser.DefaultParseValue("--flag"), "--flag")

    def testDefaultParseValueQuotedStrings(self):
        self.assertEqual(parser.DefaultParseValue("'hello'"), "hello")
        self.assertEqual(parser.DefaultParseValue("'hello world'"), "hello world")
        self.assertEqual(parser.DefaultParseValue("'--flag'"), "--flag")
        self.assertEqual(parser.DefaultParseValue('"hello"'), "hello")
        self.assertEqual(parser.DefaultParseValue('"hello world"'), "hello world")
        self.assertEqual(parser.DefaultParseValue('"--flag"'), "--flag")

    def testDefaultParseValueSpecialStrings(self):
        self.assertEqual(parser.DefaultParseValue("-"), "-")
        self.assertEqual(parser.DefaultParseValue("--"), "--")
        self.assertEqual(parser.DefaultParseValue("---"), "---")
        self.assertEqual(parser.DefaultParseValue("----"), "----")
        self.assertEqual(parser.DefaultParseValue("None"), None)
        self.assertEqual(parser.DefaultParseValue("'None'"), "None")

    def testDefaultParseValueNumbers(self):
        self.assertEqual(parser.DefaultParseValue("23"), 23)
        self.assertEqual(parser.DefaultParseValue("-23"), -23)
        self.assertEqual(parser.DefaultParseValue("23.0"), 23.0)
        self.assertIsInstance(parser.DefaultParseValue("23"), int)
        self.assertIsInstance(parser.DefaultParseValue("23.0"), float)
        self.assertEqual(parser.DefaultParseValue("23.5"), 23.5)
        self.assertEqual(parser.DefaultParseValue("-23.5"), -23.5)

    def testDefaultParseValueStringNumbers(self):
        self.assertEqual(parser.DefaultParseValue("'23'"), "23")
        self.assertEqual(parser.DefaultParseValue("'23.0'"), "23.0")
        self.assertEqual(parser.DefaultParseValue("'23.5'"), "23.5")
        self.assertEqual(parser.DefaultParseValue('"23"'), "23")
        self.assertEqual(parser.DefaultParseValue('"23.0"'), "23.0")
        self.assertEqual(parser.DefaultParseValue('"23.5"'), "23.5")

    def testDefaultParseValueQuotedStringNumbers(self):
        self.assertEqual(parser.DefaultParseValue("\"'123'\""), "'123'")

    def testDefaultParseValueOtherNumbers(self):
        self.assertEqual(parser.DefaultParseValue("1e5"), 100000.0)

    def testDefaultParseValueLists(self):
        self.assertEqual(parser.DefaultParseValue("[1, 2, 3]"), [1, 2, 3])
        self.assertEqual(parser.DefaultParseValue('[1, "2", 3]'), [1, "2", 3])
        self.assertEqual(parser.DefaultParseValue("[1, '\"2\"', 3]"), [1, '"2"', 3])
        self.assertEqual(parser.DefaultParseValue('[1, "hello", 3]'), [1, "hello", 3])

    def testDefaultParseValueBareWordsLists(self):
        self.assertEqual(parser.DefaultParseValue('[one, 2, "3"]'), ["one", 2, "3"])

    def testDefaultParseValueDict(self):
        self.assertEqual(parser.DefaultParseValue('{"abc": 5, "123": 1}'), {"abc": 5, "123": 1})

    def testDefaultParseValueNone(self):
        self.assertEqual(parser.DefaultParseValue("None"), None)

    def testDefaultParseValueBool(self):
        self.assertEqual(parser.DefaultParseValue("True"), True)
        self.assertEqual(parser.DefaultParseValue("False"), False)

    def testDefaultParseValueBareWordsTuple(self):
        self.assertEqual(parser.DefaultParseValue('(one, 2, "3")'), ("one", 2, "3"))
        self.assertEqual(parser.DefaultParseValue('one, "2", 3'), ("one", "2", 3))

    def testDefaultParseValueNestedContainers(self):
        self.assertEqual(
            parser.DefaultParseValue('[(A, 2, "3"), 5, {alpha: 10.2, beta: "cat"}]'),
            [("A", 2, "3"), 5, {"alpha": 10.2, "beta": "cat"}],
        )

    def testDefaultParseValueComments(self):
        self.assertEqual(parser.DefaultParseValue('"0#comments"'), "0#comments")
        # Comments are stripped. This behavior may change in the future.
        self.assertEqual(parser.DefaultParseValue("0#comments"), 0)

    def testDefaultParseValueBadLiteral(self):
        # If it can't be parsed, we treat it as a string. This behavior may change.
        self.assertEqual(parser.DefaultParseValue('[(A, 2, "3"), 5'), '[(A, 2, "3"), 5')
        self.assertEqual(parser.DefaultParseValue("x=10"), "x=10")

    def testDefaultParseValueSyntaxError(self):
        # If it can't be parsed, we treat it as a string.
        self.assertEqual(parser.DefaultParseValue('"'), '"')

    def testDefaultParseValueIgnoreBinOp(self):
        self.assertEqual(parser.DefaultParseValue("2017-10-10"), "2017-10-10")
        self.assertEqual(parser.DefaultParseValue("1+1"), "1+1")


if __name__ == "__main__":
    testutils.main()



# -*- coding: utf-8 -*- #
# Copyright 2013 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for determining the current platform and architecture."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import os
import platform
import subprocess
import sys


class Error(Exception):
    """Base class for exceptions in the platforms module."""

    pass


class InvalidEnumValue(Error):  # pylint: disable=g-bad-exception-name
    """Exception for when a string could not be parsed to a valid enum value."""

    def __init__(self, given, enum_type, options):
        """Constructs a new exception.

        Args:
          given: str, The given string that could not be parsed.
          enum_type: str, The human readable name of the enum you were trying to
            parse.
          options: list(str), The valid values for this enum.
        """
        super(InvalidEnumValue, self).__init__(
            "Could not parse [{0}] into a valid {1}.  Valid values are [{2}]".format(
                given, enum_type, ", ".join(options)
            )
        )


class OperatingSystem(object):
    """An enum representing the operating system you are running on."""

    class _OS(object):
        """A single operating system."""

        # pylint: disable=redefined-builtin
        def __init__(self, id, name, file_name):
            self.id = id
            self.name = name
            self.file_name = file_name

        def __str__(self):
            return self.id

        def __eq__(self, other):
            return (
                isinstance(other, type(self))
                and self.id == other.id
                and self.name == other.name
                and self.file_name == other.file_name
            )

        def __hash__(self):
            return hash(self.id) + hash(self.name) + hash(self.file_name)

        def __ne__(self, other):
            return not self == other

        @classmethod
        def _CmpHelper(cls, x, y):
            """Just a helper equivalent to the cmp() function in Python 2."""
            return (x > y) - (x < y)

        def __lt__(self, other):
            return self._CmpHelper((self.id, self.name, self.file_name), (other.id, other.name, other.file_name)) < 0

        def __gt__(self, other):
            return self._CmpHelper((self.id, self.name, self.file_name), (other.id, other.name, other.file_name)) > 0

        def __le__(self, other):
            return not self.__gt__(other)

        def __ge__(self, other):
            return not self.__lt__(other)

    WINDOWS = _OS("WINDOWS", "Windows", "windows")
    MACOSX = _OS("MACOSX", "Mac OS X", "darwin")
    LINUX = _OS("LINUX", "Linux", "linux")
    CYGWIN = _OS("CYGWIN", "Cygwin", "cygwin")
    MSYS = _OS("MSYS", "Msys", "msys")
    _ALL = [WINDOWS, MACOSX, LINUX, CYGWIN, MSYS]

    @staticmethod
    def AllValues():
        """Gets all possible enum values.

        Returns:
          list, All the enum values.
        """
        return list(OperatingSystem._ALL)

    @staticmethod
    def FromId(os_id, error_on_unknown=True):
        """Gets the enum corresponding to the given operating system id.

        Args:
          os_id: str, The operating system id to parse
          error_on_unknown: bool, True to raise an exception if the id is unknown,
            False to just return None.

        Raises:
          InvalidEnumValue: If the given value cannot be parsed.

        Returns:
          OperatingSystemTuple, One of the OperatingSystem constants or None if the
          input is None.
        """
        if not os_id:
            return None
        for operating_system in OperatingSystem._ALL:
            if operating_system.id == os_id:
                return operating_system
        if error_on_unknown:
            raise InvalidEnumValue(os_id, "Operating System", [value.id for value in OperatingSystem._ALL])
        return None

    @staticmethod
    def Current():
        """Determines the current operating system.

        Returns:
          OperatingSystemTuple, One of the OperatingSystem constants or None if it
          cannot be determined.
        """
        if os.name == "nt":
            return OperatingSystem.WINDOWS
        elif "linux" in sys.platform:
            return OperatingSystem.LINUX
        elif "darwin" in sys.platform:
            return OperatingSystem.MACOSX
        elif "cygwin" in sys.platform:
            return OperatingSystem.CYGWIN
        return None

    @staticmethod
    def IsWindows():
        """Returns True if the current operating system is Windows."""
        return OperatingSystem.Current() is OperatingSystem.WINDOWS


class Architecture(object):
    """An enum representing the system architecture you are running on."""

    class _ARCH(object):
        """A single architecture."""

        # pylint: disable=redefined-builtin
        def __init__(self, id, name, file_name):
            self.id = id
            self.name = name
            self.file_name = file_name

        def __str__(self):
            return self.id

        def __eq__(self, other):
            return (
                isinstance(other, type(self))
                and self.id == other.id
                and self.name == other.name
                and self.file_name == other.file_name
            )

        def __hash__(self):
            return hash(self.id) + hash(self.name) + hash(self.file_name)

        def __ne__(self, other):
            return not self == other

        @classmethod
        def _CmpHelper(cls, x, y):
            """Just a helper equivalent to the cmp() function in Python 2."""
            return (x > y) - (x < y)

        def __lt__(self, other):
            return self._CmpHelper((self.id, self.name, self.file_name), (other.id, other.name, other.file_name)) < 0

        def __gt__(self, other):
            return self._CmpHelper((self.id, self.name, self.file_name), (other.id, other.name, other.file_name)) > 0

        def __le__(self, other):
            return not self.__gt__(other)

        def __ge__(self, other):
            return not self.__lt__(other)

    x86 = _ARCH("x86", "x86", "x86")
    x86_64 = _ARCH("x86_64", "x86_64", "x86_64")
    ppc = _ARCH("PPC", "PPC", "ppc")
    arm = _ARCH("arm", "arm", "arm")
    _ALL = [x86, x86_64, ppc, arm]

    # Possible values for `uname -m` and what arch they map to.
    # Examples of possible values: https://en.wikipedia.org/wiki/Uname
    _MACHINE_TO_ARCHITECTURE = {
        "amd64": x86_64,
        "x86_64": x86_64,
        "i686-64": x86_64,
        "i386": x86,
        "i686": x86,
        "x86": x86,
        "ia64": x86,  # Itanium is different x64 arch, treat it as the common x86.
        "powerpc": ppc,
        "power macintosh": ppc,
        "ppc64": ppc,
        "armv6": arm,
        "armv6l": arm,
        "arm64": arm,
        "armv7": arm,
        "armv7l": arm,
    }

    @staticmethod
    def AllValues():
        """Gets all possible enum values.

        Returns:
          list, All the enum values.
        """
        return list(Architecture._ALL)

    @staticmethod
    def FromId(architecture_id, error_on_unknown=True):
        """Gets the enum corresponding to the given architecture id.

        Args:
          architecture_id: str, The architecture id to parse
          error_on_unknown: bool, True to raise an exception if the id is unknown,
            False to just return None.

        Raises:
          InvalidEnumValue: If the given value cannot be parsed.

        Returns:
          ArchitectureTuple, One of the Architecture constants or None if the input
          is None.
        """
        if not architecture_id:
            return None
        for arch in Architecture._ALL:
            if arch.id == architecture_id:
                return arch
        if error_on_unknown:
            raise InvalidEnumValue(architecture_id, "Architecture", [value.id for value in Architecture._ALL])
        return None

    @staticmethod
    def Current():
        """Determines the current system architecture.

        Returns:
          ArchitectureTuple, One of the Architecture constants or None if it cannot
          be determined.
        """
        return Architecture._MACHINE_TO_ARCHITECTURE.get(platform.machine().lower())


class Platform(object):
    """Holds an operating system and architecture."""

    def __init__(self, operating_system, architecture):
        """Constructs a new platform.

        Args:
          operating_system: OperatingSystem, The OS
          architecture: Architecture, The machine architecture.
        """
        self.operating_system = operating_system
        self.architecture = architecture

    def __str__(self):
        return "{}-{}".format(self.operating_system, self.architecture)

    @staticmethod
    def Current(os_override=None, arch_override=None):
        """Determines the current platform you are running on.

        Args:
          os_override: OperatingSystem, A value to use instead of the current.
          arch_override: Architecture, A value to use instead of the current.

        Returns:
          Platform, The platform tuple of operating system and architecture.  Either
          can be None if it could not be determined.
        """
        return Platform(
            os_override if os_override else OperatingSystem.Current(),
            arch_override if arch_override else Architecture.Current(),
        )

    def UserAgentFragment(self):
        """Generates the fragment of the User-Agent that represents the OS.

        Examples:
          (Linux 3.2.5-gg1236)
          (Windows NT 6.1.7601)
          (Macintosh; PPC Mac OS X 12.4.0)
          (Macintosh; Intel Mac OS X 12.4.0)

        Returns:
          str, The fragment of the User-Agent string.
        """
        # Below, there are examples of the value of platform.uname() per platform.
        # platform.release() is uname[2], platform.version() is uname[3].
        if self.operating_system == OperatingSystem.LINUX:
            # ('Linux', '<hostname goes here>', '3.2.5-gg1236',
            # '#1 SMP Tue May 21 02:35:06 PDT 2013', 'x86_64', 'x86_64')
            return "({name} {version})".format(name=self.operating_system.name, version=platform.release())
        elif self.operating_system == OperatingSystem.WINDOWS:
            # ('Windows', '<hostname goes here>', '7', '6.1.7601', 'AMD64',
            # 'Intel64 Family 6 Model 45 Stepping 7, GenuineIntel')
            return "({name} NT {version})".format(name=self.operating_system.name, version=platform.version())
        elif self.operating_system == OperatingSystem.MACOSX:
            # ('Darwin', '<hostname goes here>', '12.4.0',
            # 'Darwin Kernel Version 12.4.0: Wed May  1 17:57:12 PDT 2013;
            # root:xnu-2050.24.15~1/RELEASE_X86_64', 'x86_64', 'i386')
            format_string = "(Macintosh; {name} Mac OS X {version})"
            arch_string = self.architecture.name if self.architecture == Architecture.ppc else "Intel"
            return format_string.format(name=arch_string, version=platform.release())
        else:
            return "()"

    def AsyncPopenArgs(self):
        """Returns the args for spawning an async process using Popen on this OS.

        Make sure the main process does not wait for the new process. On windows
        this means setting the 0x8 creation flag to detach the process.

        Killing a group leader kills the whole group. Setting creation flag 0x200 on
        Windows or running setsid on *nix makes sure the new process is in a new
        session with the new process the group leader. This means it can't be killed
        if the parent is killed.

        Finally, all file descriptors (FD) need to be closed so that waiting for the
        output of the main process does not inadvertently wait for the output of the
        new process, which means waiting for the termination of the new process.
        If the new process wants to write to a file, it can open new FDs.

        Returns:
          {str:}, The args for spawning an async process using Popen on this OS.
        """
        args = {}
        if self.operating_system == OperatingSystem.WINDOWS:
            args["close_fds"] = True  # This is enough to close _all_ FDs on windows.
            detached_process = 0x00000008
            create_new_process_group = 0x00000200
            # 0x008 | 0x200 == 0x208
            args["creationflags"] = detached_process | create_new_process_group
        else:
            # Killing a group leader kills the whole group.
            # Create a new session with the new process the group leader.
            args["preexec_fn"] = os.setsid
            args["close_fds"] = True  # This closes all FDs _except_ 0, 1, 2 on *nix.
            args["stdin"] = subprocess.PIPE
            args["stdout"] = subprocess.PIPE
            args["stderr"] = subprocess.PIPE
        return args


class PythonVersion(object):
    """Class to validate the Python version we are using.

    The Cloud SDK officially supports Python 2.7.

    However, many commands do work with Python 2.6, so we don't error out when
    users are using this (we consider it sometimes "compatible" but not
    "supported").
    """

    # See class docstring for descriptions of what these mean
    MIN_REQUIRED_PY2_VERSION = (2, 6)
    MIN_SUPPORTED_PY2_VERSION = (2, 7)
    MIN_SUPPORTED_PY3_VERSION = (3, 4)
    ENV_VAR_MESSAGE = """\

If you have a compatible Python interpreter installed, you can use it by setting
the CLOUDSDK_PYTHON environment variable to point to it.

"""

    def __init__(self, version=None):
        if version:
            self.version = version
        elif hasattr(sys, "version_info"):
            self.version = sys.version_info[:2]
        else:
            self.version = None

    def SupportedVersionMessage(self, allow_py3):
        if allow_py3:
            return "Please use Python version {0}.{1}.x or {2}.{3} and up.".format(
                PythonVersion.MIN_SUPPORTED_PY2_VERSION[0],
                PythonVersion.MIN_SUPPORTED_PY2_VERSION[1],
                PythonVersion.MIN_SUPPORTED_PY3_VERSION[0],
                PythonVersion.MIN_SUPPORTED_PY3_VERSION[1],
            )
        else:
            return "Please use Python version {0}.{1}.x.".format(
                PythonVersion.MIN_SUPPORTED_PY2_VERSION[0], PythonVersion.MIN_SUPPORTED_PY2_VERSION[1]
            )

    def IsCompatible(self, allow_py3=False, raise_exception=False):
        """Ensure that the Python version we are using is compatible.

        This will print an error message if not compatible.

        Compatible versions are 2.6 and 2.7 and > 3.4 if allow_py3 is True.
        We don't guarantee support for 2.6 so we want to warn about it.

        Args:
          allow_py3: bool, True if we should allow a Python 3 interpreter to run
            gcloud. If False, this returns an error for Python 3.
          raise_exception: bool, True to raise an exception rather than printing
            the error and exiting.

        Raises:
          Error: If not compatible and raise_exception is True.

        Returns:
          bool, True if the version is valid, False otherwise.
        """
        error = None
        if not self.version:
            # We don't know the version, not a good sign.
            error = (
                "ERROR: Your current version of Python is not compatible with "
                "the Google Cloud SDK. {0}\n".format(self.SupportedVersionMessage(allow_py3))
            )
        else:
            if self.version[0] < 3:
                # Python 2 Mode
                if self.version < PythonVersion.MIN_REQUIRED_PY2_VERSION:
                    error = "ERROR: Python {0}.{1} is not compatible with the Google " "Cloud SDK. {2}\n".format(
                        self.version[0], self.version[1], self.SupportedVersionMessage(allow_py3)
                    )
            else:
                # Python 3 Mode
                if not allow_py3:
                    error = "ERROR: Python 3 and later is not compatible with the " "Google Cloud SDK. {0}\n".format(
                        self.SupportedVersionMessage(allow_py3)
                    )
                elif self.version < PythonVersion.MIN_SUPPORTED_PY3_VERSION:
                    error = "ERROR: Python {0}.{1} is not compatible with the Google " "Cloud SDK. {2}\n".format(
                        self.version[0], self.version[1], self.SupportedVersionMessage(allow_py3)
                    )

        if error:
            if raise_exception:
                raise Error(error)
            sys.stderr.write(error)
            sys.stderr.write(PythonVersion.ENV_VAR_MESSAGE)
            return False

        # Warn that 2.6 might not work.
        if self.version >= self.MIN_REQUIRED_PY2_VERSION and self.version < self.MIN_SUPPORTED_PY2_VERSION:
            sys.stderr.write(
                """\
WARNING:  Python 2.6.x is no longer officially supported by the Google Cloud SDK
and may not function correctly.  {0}
{1}""".format(
                    self.SupportedVersionMessage(allow_py3), PythonVersion.ENV_VAR_MESSAGE
                )
            )

        return True



# Copyright (C) 2018 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the core module."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from afire import core
from afire import test_components as tc
from afire import testutils
from afire import trace
import mock

import six


class CoreTest(testutils.BaseTestCase):
    def testOneLineResult(self):
        self.assertEqual(core._OneLineResult(1), "1")  # pylint: disable=protected-access
        self.assertEqual(core._OneLineResult("hello"), "hello")  # pylint: disable=protected-access
        self.assertEqual(core._OneLineResult({}), "{}")  # pylint: disable=protected-access
        self.assertEqual(core._OneLineResult({"x": "y"}), '{"x": "y"}')  # pylint: disable=protected-access

    def testOneLineResultCircularRef(self):
        circular_reference = tc.CircularReference()
        self.assertEqual(
            core._OneLineResult(circular_reference.create()), "{'y': {...}}"  # pylint: disable=protected-access
        )

    @mock.patch("afire.interact.Embed")
    def testInteractiveMode(self, mock_embed):
        core.Fire(tc.TypedProperties, command=["alpha"])
        self.assertFalse(mock_embed.called)
        core.Fire(tc.TypedProperties, command=["alpha", "--", "-i"])
        self.assertTrue(mock_embed.called)

    @mock.patch("afire.interact.Embed")
    def testInteractiveModeFullArgument(self, mock_embed):
        core.Fire(tc.TypedProperties, command=["alpha", "--", "--interactive"])
        self.assertTrue(mock_embed.called)

    @mock.patch("afire.interact.Embed")
    def testInteractiveModeVariables(self, mock_embed):
        core.Fire(tc.WithDefaults, command=["double", "2", "--", "-i"])
        self.assertTrue(mock_embed.called)
        (variables, verbose), unused_kwargs = mock_embed.call_args
        self.assertFalse(verbose)
        self.assertEqual(variables["result"], 4)
        self.assertIsInstance(variables["self"], tc.WithDefaults)
        self.assertIsInstance(variables["trace"], trace.FireTrace)

    @mock.patch("afire.interact.Embed")
    def testInteractiveModeVariablesWithName(self, mock_embed):
        core.Fire(tc.WithDefaults, command=["double", "2", "--", "-i", "-v"], name="D")
        self.assertTrue(mock_embed.called)
        (variables, verbose), unused_kwargs = mock_embed.call_args
        self.assertTrue(verbose)
        self.assertEqual(variables["result"], 4)
        self.assertIsInstance(variables["self"], tc.WithDefaults)
        self.assertEqual(variables["D"], tc.WithDefaults)
        self.assertIsInstance(variables["trace"], trace.FireTrace)

    # TODO(dbieber): Use parameterized tests to break up repetitive tests.
    def testHelpWithClass(self):
        with self.assertRaisesFireExit(0, "SYNOPSIS.*ARG1"):
            core.Fire(tc.InstanceVars, command=["--", "--help"])
        with self.assertRaisesFireExit(0, "INFO:.*SYNOPSIS.*ARG1"):
            core.Fire(tc.InstanceVars, command=["--help"])
        with self.assertRaisesFireExit(0, "INFO:.*SYNOPSIS.*ARG1"):
            core.Fire(tc.InstanceVars, command=["-h"])

    def testHelpWithMember(self):
        with self.assertRaisesFireExit(0, "SYNOPSIS.*capitalize"):
            core.Fire(tc.TypedProperties, command=["gamma", "--", "--help"])
        with self.assertRaisesFireExit(0, "INFO:.*SYNOPSIS.*capitalize"):
            core.Fire(tc.TypedProperties, command=["gamma", "--help"])
        with self.assertRaisesFireExit(0, "INFO:.*SYNOPSIS.*capitalize"):
            core.Fire(tc.TypedProperties, command=["gamma", "-h"])
        with self.assertRaisesFireExit(0, "INFO:.*SYNOPSIS.*delta"):
            core.Fire(tc.TypedProperties, command=["delta", "--help"])
        with self.assertRaisesFireExit(0, "INFO:.*SYNOPSIS.*echo"):
            core.Fire(tc.TypedProperties, command=["echo", "--help"])

    def testHelpOnErrorInConstructor(self):
        with self.assertRaisesFireExit(0, "SYNOPSIS.*VALUE"):
            core.Fire(tc.ErrorInConstructor, command=["--", "--help"])
        with self.assertRaisesFireExit(0, "INFO:.*SYNOPSIS.*VALUE"):
            core.Fire(tc.ErrorInConstructor, command=["--help"])

    def testHelpWithNamespaceCollision(self):
        # Tests cases when calling the help shortcut should not show help.
        with self.assertOutputMatches(stdout="DESCRIPTION.*", stderr=None):
            core.Fire(tc.WithHelpArg, command=["--help", "False"])
        with self.assertOutputMatches(stdout="help in a dict", stderr=None):
            core.Fire(tc.WithHelpArg, command=["dictionary", "__help"])
        with self.assertOutputMatches(stdout="{}", stderr=None):
            core.Fire(tc.WithHelpArg, command=["dictionary", "--help"])
        with self.assertOutputMatches(stdout="False", stderr=None):
            core.Fire(tc.function_with_help, command=["False"])

    def testInvalidParameterRaisesFireExit(self):
        with self.assertRaisesFireExit(2, "runmisspelled"):
            core.Fire(tc.Kwargs, command=["props", "--a=1", "--b=2", "runmisspelled"])

    def testErrorRaising(self):
        # Errors in user code should not be caught; they should surface as normal.
        # This will lead to exit status code 1 for the client program.
        with self.assertRaises(ValueError):
            core.Fire(tc.ErrorRaiser, command=["fail"])

    def testFireError(self):
        error = core.FireError("Example error")
        self.assertIsNotNone(error)

    def testFireErrorMultipleValues(self):
        error = core.FireError("Example error", "value")
        self.assertIsNotNone(error)

    def testPrintEmptyDict(self):
        with self.assertOutputMatches(stdout="{}", stderr=None):
            core.Fire(tc.EmptyDictOutput, command=["totally_empty"])
        with self.assertOutputMatches(stdout="{}", stderr=None):
            core.Fire(tc.EmptyDictOutput, command=["nothing_printable"])

    def testPrintOrderedDict(self):
        with self.assertOutputMatches(stdout=r"A:\s+A\s+2:\s+2\s+", stderr=None):
            core.Fire(tc.OrderedDictionary, command=["non_empty"])
        with self.assertOutputMatches(stdout="{}"):
            core.Fire(tc.OrderedDictionary, command=["empty"])

    def testPrintNamedTupleField(self):
        with self.assertOutputMatches(stdout="11", stderr=None):
            core.Fire(tc.NamedTuple, command=["point", "x"])

    def testPrintNamedTupleFieldNameEqualsValue(self):
        with self.assertOutputMatches(stdout="x", stderr=None):
            core.Fire(tc.NamedTuple, command=["matching_names", "x"])

    def testPrintNamedTupleIndex(self):
        with self.assertOutputMatches(stdout="22", stderr=None):
            core.Fire(tc.NamedTuple, command=["point", "1"])

    def testPrintSet(self):
        with self.assertOutputMatches(stdout=".*three.*", stderr=None):
            core.Fire(tc.simple_set(), command=[])

    def testPrintFrozenSet(self):
        with self.assertOutputMatches(stdout=".*three.*", stderr=None):
            core.Fire(tc.simple_frozenset(), command=[])

    def testPrintNamedTupleNegativeIndex(self):
        with self.assertOutputMatches(stdout="11", stderr=None):
            core.Fire(tc.NamedTuple, command=["point", "-2"])

    def testTypedCallable(self):
        with self.assertOutputMatches(stdout=r"1 int foo", stderr=None):
            core.Fire(tc.CallableWithTypedKeywordArgument().IntType, command=["--foo=1"])

        with self.assertOutputMatches(stdout=r"1 str foo", stderr=None):
            core.Fire(tc.CallableWithTypedKeywordArgument().StrType, command=["--foo=1"])

        with self.assertOutputMatches(stdout=r"None str foo", stderr=None):
            core.Fire(tc.CallableWithTypedKeywordArgument().StrType, command=["--foo=None"])

        with self.assertOutputMatches(stdout=r"2023-09-24 12:52:33 datetime foo", stderr=None):
            core.Fire(tc.CallableWithTypedKeywordArgument().DatetimeType, command=["--foo=2023-09-24 12:52:33"])

        # complex type
        with self.assertOutputMatches(stdout=r"{'1': 2} dict foo", stderr=None):
            core.Fire(tc.CallableWithTypedKeywordArgument().DictType, command=["--foo={1:2}"])

        with self.assertOutputMatches(
            stdout=r"{'1': {3: datetime.datetime\(2023, 9, 24, 12, 52, 33\)}} dict foo", stderr=None
        ):
            core.Fire(
                tc.CallableWithTypedKeywordArgument().DictInDictType, command=["--foo={1:{3:'2023-09-24 12:52:33'}}"]
            )

        with self.assertOutputMatches(stdout=r"1 int foo", stderr=None):
            core.Fire(tc.CallableWithTypedKeywordArgument().UnionType, command=["--foo=1"])

        with self.assertOutputMatches(stdout=r"2023-09-24 12:52:33 datetime foo", stderr=None):
            core.Fire(tc.CallableWithTypedKeywordArgument().UnionType, command=["--foo=2023-09-24 12:52:33"])

        with self.assertOutputMatches(stdout=r"abcdef str foo", stderr=None):
            core.Fire(tc.CallableWithTypedKeywordArgument().UnionType, command=["--foo=abcdef"])

        with self.assertOutputMatches(stdout=r"None NoneType foo", stderr=None):
            core.Fire(tc.CallableWithTypedKeywordArgument().OptionalType, command=["--foo=None"])

        with self.assertOutputMatches(stdout=r"2023-09-24 12:52:33 datetime foo", stderr=None):
            core.Fire(tc.CallableWithTypedKeywordArgument().OptionalType, command=["--foo=2023-09-24 12:52:33"])

        with self.assertOutputMatches(stdout=r"None NoneType foo", stderr=None):
            core.Fire(tc.CallableWithTypedKeywordArgument().DictOptionalUnion, command=["--foo=None"])

        with self.assertOutputMatches(stdout=r"{xxx:yyy} str foo", stderr=None):
            core.Fire(tc.CallableWithTypedKeywordArgument().DictOptionalUnion, command=["--foo={xxx:yyy}"])

        with self.assertOutputMatches(stdout=r"{'xxx': 1} dict foo", stderr=None):
            core.Fire(tc.CallableWithTypedKeywordArgument().DictOptionalUnion, command=["--foo={xxx:1}"])

        with self.assertOutputMatches(stdout=r"{xxx:yyy} str foo", stderr=None):
            core.Fire(tc.CallableWithTypedKeywordArgument().DictOptionalUnion, command=["--foo={xxx:yyy}"])

        with self.assertOutputMatches(stdout=r"{'xxx': None} dict foo", stderr=None):
            core.Fire(tc.CallableWithTypedKeywordArgument().DictOptionalUnion, command=["--foo={xxx:None}"])

        with self.assertOutputMatches(stdout=r"\(1, '2'\) tuple foo", stderr=None):
            core.Fire(tc.CallableWithTypedKeywordArgument().TupleType, command=["--foo=(1, 2)"])

        with self.assertOutputMatches(stdout=r"\(1, '2'\) tuple foo", stderr=None):
            core.Fire(tc.CallableWithTypedKeywordArgument().TupleOptionalType, command=["--foo=(1, 2)"])

        with self.assertOutputMatches(stdout=r"\(1, None\) tuple foo", stderr=None):
            core.Fire(tc.CallableWithTypedKeywordArgument().TupleOptionalType, command=["--foo=(1, None)"])

        with self.assertOutputMatches(stdout=r"\['1', '2'\] list foo", stderr=None):
            core.Fire(tc.CallableWithTypedKeywordArgument().ListType, command=["--foo=[1, 2]"])

        with self.assertOutputMatches(stdout=r"\{'[1,2]', '[1,2]'\} set foo", stderr=None):
            core.Fire(tc.CallableWithTypedKeywordArgument().SetType, command=["--foo={1, 2}"])

        with self.assertOutputMatches(stdout=r"b'xyz' bytes foo", stderr=None):
            core.Fire(tc.CallableWithTypedKeywordArgument().BytesType, command=["--foo=b'xyz'"])

        with self.assertOutputMatches(stdout=r"b'xyz' bytes foo", stderr=None):
            core.Fire(tc.CallableWithTypedKeywordArgument().BytesType, command=["--foo=xyz"])

        with self.assertOutputMatches(
            stdout=r"\['1', datetime.datetime\(2023, 9, 24, 12, 52, 33\)\] list foo", stderr=None
        ):
            core.Fire(tc.CallableWithTypedKeywordArgument().ListUnionType, command=["--foo=[1, '2023-09-24 12:52:33']"])

    def testCallable(self):
        with self.assertOutputMatches(stdout=r"foo:\s+foo\s+", stderr=None):
            core.Fire(tc.CallableWithKeywordArgument(), command=["--foo=foo"])
        with self.assertOutputMatches(stdout=r"foo\s+", stderr=None):
            core.Fire(tc.CallableWithKeywordArgument(), command=["print_msg", "foo"])
        with self.assertOutputMatches(stdout=r"", stderr=None):
            core.Fire(tc.CallableWithKeywordArgument(), command=[])

    def testCallableWithPositionalArgs(self):
        with self.assertRaisesFireExit(2, ""):
            # This does not give 7 since positional args are disallowed for callable
            # objects.
            core.Fire(tc.CallableWithPositionalArgs(), command=["3", "4"])

    def testStaticMethod(self):
        self.assertEqual(
            core.Fire(tc.HasStaticAndClassMethods, command=["static_fn", "alpha"]),
            "alpha",
        )

    def testClassMethod(self):
        self.assertEqual(
            core.Fire(tc.HasStaticAndClassMethods, command=["class_fn", "6"]),
            7,
        )

    def testCustomSerialize(self):
        def serialize(x):
            if isinstance(x, list):
                return ", ".join(str(xi) for xi in x)
            if isinstance(x, dict):
                return ", ".join("{}={!r}".format(k, v) for k, v in sorted(x.items()))
            if x == "special":
                return ["SURPRISE!!", "I'm a list!"]
            return x

        ident = lambda x: x

        with self.assertOutputMatches(stdout="a, b", stderr=None):
            _ = core.Fire(ident, command=["[a,b]"], serialize=serialize)
        with self.assertOutputMatches(stdout="a=5, b=6", stderr=None):
            _ = core.Fire(ident, command=["{a:5,b:6}"], serialize=serialize)
        with self.assertOutputMatches(stdout="asdf", stderr=None):
            _ = core.Fire(ident, command=["asdf"], serialize=serialize)
        with self.assertOutputMatches(stdout="SURPRISE!!\nI'm a list!\n", stderr=None):
            _ = core.Fire(ident, command=["special"], serialize=serialize)
        with self.assertRaises(core.FireError):
            core.Fire(ident, command=["asdf"], serialize=55)

    @testutils.skipIf(six.PY2, "lru_cache is Python 3 only.")
    def testLruCacheDecoratorBoundArg(self):
        self.assertEqual(
            core.Fire(
                tc.py3.LruCacheDecoratedMethod, command=["lru_cache_in_class", "foo"]  # pytype: disable=module-attr
            ),
            "foo",
        )

    @testutils.skipIf(six.PY2, "lru_cache is Python 3 only.")
    def testLruCacheDecorator(self):
        self.assertEqual(core.Fire(tc.py3.lru_cache_decorated, command=["foo"]), "foo")  # pytype: disable=module-attr


if __name__ == "__main__":
    testutils.main()



# Copyright (C) 2018 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provides tab completion functionality for CLIs built with Fire."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import copy
import inspect

from afire import inspectutils
import six


def Script(name, component, default_options=None, shell="bash"):
    if shell == "fish":
        return _FishScript(name, _Commands(component), default_options)
    return _BashScript(name, _Commands(component), default_options)


def _BashScript(name, commands, default_options=None):
    """Returns a Bash script registering a completion function for the commands.

    Args:
      name: The first token in the commands, also the name of the command.
      commands: A list of all possible commands that tab completion can complete
          to. Each command is a list or tuple of the string tokens that make up
          that command.
      default_options: A dict of options that can be used with any command. Use
          this if there are flags that can always be appended to a command.
    Returns:
      A string which is the Bash script. Source the bash script to enable tab
      completion in Bash.
    """
    default_options = default_options or set()
    global_options, options_map, subcommands_map = _GetMaps(name, commands, default_options)

    bash_completion_template = """# bash completion support for {name}
# DO NOT EDIT.
# This script is autogenerated by fire/completion.py.

_complete-{identifier}()
{{
  local cur prev opts lastcommand
  COMPREPLY=()
  prev="${{COMP_WORDS[COMP_CWORD-1]}}"
  cur="${{COMP_WORDS[COMP_CWORD]}}"
  lastcommand=$(get_lastcommand)

  opts="{default_options}"
  GLOBAL_OPTIONS="{global_options}"

{checks}

  COMPREPLY=( $(compgen -W "${{opts}}" -- ${{cur}}) )
  return 0
}}

get_lastcommand()
{{
  local lastcommand i

  lastcommand=
  for ((i=0; i < ${{#COMP_WORDS[@]}}; ++i)); do
    if [[ ${{COMP_WORDS[i]}} != -* ]] && [[ -n ${{COMP_WORDS[i]}} ]] && [[
      ${{COMP_WORDS[i]}} != $cur ]]; then
      lastcommand=${{COMP_WORDS[i]}}
    fi
  done

  echo $lastcommand
}}

filter_options()
{{
  local opts
  opts=""
  for opt in "$@"
  do
    if ! option_already_entered $opt; then
      opts="$opts $opt"
    fi
  done

  echo $opts
}}

option_already_entered()
{{
  local opt
  for opt in ${{COMP_WORDS[@]:0:COMP_CWORD}}
  do
    if [ $1 == $opt ]; then
      return 0
    fi
  done
  return 1
}}

is_prev_global()
{{
  local opt
  for opt in $GLOBAL_OPTIONS
  do
    if [ $opt == $prev ]; then
      return 0
    fi
  done
  return 1
}}

complete -F _complete-{identifier} {command}
"""

    check_wrapper = """
  case "${{lastcommand}}" in
  {lastcommand_checks}
  esac"""

    lastcommand_check_template = """
    {command})
      {opts_assignment}
      opts=$(filter_options $opts)
    ;;"""

    opts_assignment_subcommand_template = """
      if is_prev_global; then
        opts="${{GLOBAL_OPTIONS}}"
      else
        opts="{options} ${{GLOBAL_OPTIONS}}"
      fi"""

    opts_assignment_main_command_template = """
      opts="{options} ${{GLOBAL_OPTIONS}}" """

    def _GetOptsAssignmentTemplate(command):
        if command == name:
            return opts_assignment_main_command_template
        else:
            return opts_assignment_subcommand_template

    lines = []
    commands_set = set()
    commands_set.add(name)
    commands_set = commands_set.union(set(subcommands_map.keys()))
    commands_set = commands_set.union(set(options_map.keys()))
    for command in commands_set:
        opts_assignment = _GetOptsAssignmentTemplate(command).format(
            options=" ".join(sorted(options_map[command].union(subcommands_map[command]))),
        )
        lines.append(lastcommand_check_template.format(command=command, opts_assignment=opts_assignment))
    lastcommand_checks = "\n".join(lines)

    checks = check_wrapper.format(
        lastcommand_checks=lastcommand_checks,
    )

    return bash_completion_template.format(
        name=name,
        command=name,
        checks=checks,
        default_options=" ".join(default_options),
        identifier=name.replace("/", "").replace(".", "").replace(",", ""),
        global_options=" ".join(global_options),
    )


def _FishScript(name, commands, default_options=None):
    """Returns a Fish script registering a completion function for the commands.

    Args:
      name: The first token in the commands, also the name of the command.
      commands: A list of all possible commands that tab completion can complete
          to. Each command is a list or tuple of the string tokens that make up
          that command.
      default_options: A dict of options that can be used with any command. Use
          this if there are flags that can always be appended to a command.
    Returns:
      A string which is the Fish script. Source the fish script to enable tab
      completion in Fish.
    """
    default_options = default_options or set()
    global_options, options_map, subcommands_map = _GetMaps(name, commands, default_options)

    fish_source = """function __fish_using_command
    set cmd (commandline -opc)
    for i in (seq (count $cmd) 1)
        switch $cmd[$i]
        case "-*"
        case "*"
            if [ $cmd[$i] = $argv[1] ]
                return 0
            else
                return 1
            end
        end
    end
    return 1
end

function __option_entered_check
    set cmd (commandline -opc)
    for i in (seq (count $cmd))
        switch $cmd[$i]
        case "-*"
            if [ $cmd[$i] = $argv[1] ]
                return 1
            end
        end
    end
    return 0
end

function __is_prev_global
    set cmd (commandline -opc)
    set global_options {global_options}
    set prev (count $cmd)

    for opt in $global_options
        if [ "--$opt" = $cmd[$prev] ]
            echo $prev
            return 0
        end
    end
    return 1
end

"""

    subcommand_template = "complete -c {name} -n '__fish_using_command " "{command}' -f -a {subcommand}\n"
    flag_template = (
        "complete -c {name} -n "
        "'__fish_using_command {command};{prev_global_check} and "
        "__option_entered_check --{option}' -l {option}\n"
    )

    prev_global_check = " and __is_prev_global;"
    for command in set(subcommands_map.keys()).union(set(options_map.keys())):
        for subcommand in subcommands_map[command]:
            fish_source += subcommand_template.format(
                name=name,
                command=command,
                subcommand=subcommand,
            )

        for option in options_map[command].union(global_options):
            check_needed = command != name
            fish_source += flag_template.format(
                name=name,
                command=command,
                prev_global_check=prev_global_check if check_needed else "",
                option=option.lstrip("--"),
            )

    return fish_source.format(global_options=" ".join('"{option}"'.format(option=option) for option in global_options))


def MemberVisible(component, name, member, class_attrs=None, verbose=False):
    """Returns whether a member should be included in auto-completion or help.

    Determines whether a member of an object with the specified name should be
    included in auto-completion or help text(both usage and detailed help).

    If the member name starts with '__', it will always be excluded. If it
    starts with only one '_', it will be included for all non-string types. If
    verbose is True, the members, including the private members, are included.

    When not in verbose mode, some modules and functions are excluded as well.

    Args:
      component: The component containing the member.
      name: The name of the member.
      member: The member itself.
      class_attrs: (optional) If component is a class, provide this as:
        inspectutils.GetClassAttrsDict(component). If not provided, it will be
        computed.
      verbose: Whether to include private members.
    Returns
      A boolean value indicating whether the member should be included.
    """
    if isinstance(name, six.string_types) and name.startswith("__"):
        return False
    if verbose:
        return True
    if member is absolute_import or member is division or member is print_function:
        return False
    if isinstance(member, type(absolute_import)) and six.PY34:
        return False
    if inspect.ismodule(member) and member is six:
        # TODO(dbieber): Determine more generally which modules to hide.
        return False
    if inspect.isclass(component):
        # If class_attrs has not been provided, compute it.
        if class_attrs is None:
            class_attrs = inspectutils.GetClassAttrsDict(class_attrs) or {}
        class_attr = class_attrs.get(name)
        if class_attr:
            # Methods and properties should only be accessible on instantiated
            # objects, not on uninstantiated classes.
            if class_attr.kind in ("method", "property"):
                return False
            # Backward compatibility notes: Before Python 3.8, namedtuple attributes
            # were properties. In Python 3.8, they have type tuplegetter.
            tuplegetter = getattr(collections, "_tuplegetter", type(None))
            if isinstance(class_attr.object, tuplegetter):
                return False
    if (
        six.PY2
        and inspect.isfunction(component)
        and name in ("func_closure", "func_code", "func_defaults", "func_dict", "func_doc", "func_globals", "func_name")
    ):
        return False
    if six.PY2 and inspect.ismethod(component) and name in ("im_class", "im_func", "im_self"):
        return False
    if isinstance(name, six.string_types):
        return not name.startswith("_")
    return True  # Default to including the member


def VisibleMembers(component, class_attrs=None, verbose=False):
    """Returns a list of the members of the given component.

    If verbose is True, then members starting with _ (normally ignored) are
    included.

    Args:
      component: The component whose members to list.
      class_attrs: (optional) If component is a class, you may provide this as:
        inspectutils.GetClassAttrsDict(component). If not provided, it will be
        computed. If provided, this determines how class members will be treated
        for visibility. In particular, methods are generally hidden for
        non-instantiated classes, but if you wish them to be shown (e.g. for
        completion scripts) then pass in a different class_attr for them.
      verbose: Whether to include private members.
    Returns:
      A list of tuples (member_name, member) of all members of the component.
    """
    if isinstance(component, dict):
        members = component.items()
    else:
        members = inspect.getmembers(component)

    # If class_attrs has not been provided, compute it.
    if class_attrs is None:
        class_attrs = inspectutils.GetClassAttrsDict(component)
    return [
        (member_name, member)
        for member_name, member in members
        if MemberVisible(component, member_name, member, class_attrs=class_attrs, verbose=verbose)
    ]


def _CompletionsFromArgs(fn_args):
    """Takes a list of fn args and returns a list of the fn's completion strings.

    Args:
      fn_args: A list of the args accepted by a function.
    Returns:
      A list of possible completion strings for that function.
    """
    completions = []
    for arg in fn_args:
        arg = arg.replace("_", "-")
        completions.append("--{arg}".format(arg=arg))
    return completions


def Completions(component, verbose=False):
    """Gives possible Fire command completions for the component.

    A completion is a string that can be appended to a command to continue that
    command. These are used for TAB-completions in Bash for Fire CLIs.

    Args:
      component: The component whose completions to list.
      verbose: Whether to include all completions, even private members.
    Returns:
      A list of completions for a command that would so far return the component.
    """
    if inspect.isroutine(component) or inspect.isclass(component):
        spec = inspectutils.GetFullArgSpec(component)
        return _CompletionsFromArgs(spec.args + spec.kwonlyargs)

    if isinstance(component, (tuple, list)):
        return [str(index) for index in range(len(component))]

    if inspect.isgenerator(component):
        # TODO(dbieber): There are currently no commands available for generators.
        return []

    return [_FormatForCommand(member_name) for member_name, _ in VisibleMembers(component, verbose=verbose)]


def _FormatForCommand(token):
    """Replaces underscores with hyphens, unless the token starts with a token.

    This is because we typically prefer hyphens to underscores at the command
    line, but we reserve hyphens at the start of a token for flags. This becomes
    relevant when --verbose is activated, so that things like __str__ don't get
    transformed into --str--, which would get confused for a flag.

    Args:
      token: The token to transform.
    Returns:
      The transformed token.
    """
    if not isinstance(token, six.string_types):
        token = str(token)

    if token.startswith("_"):
        return token

    return token.replace("_", "-")


def _Commands(component, depth=3):
    """Yields tuples representing commands.

    To use the command from Python, insert '.' between each element of the tuple.
    To use the command from the command line, insert ' ' between each element of
    the tuple.

    Args:
      component: The component considered to be the root of the yielded commands.
      depth: The maximum depth with which to traverse the member DAG for commands.
    Yields:
      Tuples, each tuple representing one possible command for this CLI.
      Only traverses the member DAG up to a depth of depth.
    """
    if inspect.isroutine(component) or inspect.isclass(component):
        for completion in Completions(component, verbose=False):
            yield (completion,)
    if inspect.isroutine(component):
        return  # Don't descend into routines.

    if depth < 1:
        return

    # By setting class_attrs={} we don't hide methods in completion.
    for member_name, member in VisibleMembers(component, class_attrs={}, verbose=False):
        # TODO(dbieber): Also skip components we've already seen.
        member_name = _FormatForCommand(member_name)

        yield (member_name,)

        for command in _Commands(member, depth - 1):
            yield (member_name,) + command


def _IsOption(arg):
    return arg.startswith("-")


def _GetMaps(name, commands, default_options):
    """Returns sets of subcommands and options for each command.

    Args:
      name: The first token in the commands, also the name of the command.
      commands: A list of all possible commands that tab completion can complete
          to. Each command is a list or tuple of the string tokens that make up
          that command.
      default_options: A dict of options that can be used with any command. Use
          this if there are flags that can always be appended to a command.
    Returns:
      global_options: A set of all options of the first token of the command.
      subcommands_map: A dict storing set of subcommands for each
          command/subcommand.
      options_map: A dict storing set of options for each subcommand.
    """
    global_options = copy.copy(default_options)
    options_map = collections.defaultdict(lambda: copy.copy(default_options))
    subcommands_map = collections.defaultdict(set)

    for command in commands:
        if len(command) == 1:
            if _IsOption(command[0]):
                global_options.add(command[0])
            else:
                subcommands_map[name].add(command[0])

        elif command:
            subcommand = command[-2]
            arg = _FormatForCommand(command[-1])

            if _IsOption(arg):
                args_map = options_map
            else:
                args_map = subcommands_map

            args_map[subcommand].add(arg)
            args_map[subcommand.replace("_", "-")].add(arg)

    return global_options, options_map, subcommands_map



# Copyright (C) 2018 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module has components that are used for testing Python Fire."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import enum
import functools
from datetime import datetime
from typing import Dict, Union, Optional, Tuple, List, Set

import six

if six.PY3:
    from afire import test_components_py3 as py3  # pylint: disable=unused-import,no-name-in-module,g-import-not-at-top


def identity(arg1, arg2, arg3=10, arg4=20, *arg5, **arg6):  # pylint: disable=keyword-arg-before-vararg
    return arg1, arg2, arg3, arg4, arg5, arg6


identity.__annotations__ = {"arg2": int, "arg4": int}


def multiplier_with_docstring(num, rate=2):
    """Multiplies num by rate.

    Args:
      num (int): the num you want to multiply
      rate (int): the rate for multiplication
    Returns:
      Multiplication of num by rate
    """
    return num * rate


def function_with_help(help=True):  # pylint: disable=redefined-builtin
    return help


class Empty(object):
    pass


class OldStyleEmpty:  # pylint: disable=old-style-class,no-init
    pass


class WithInit(object):
    def __init__(self):
        pass


class ErrorInConstructor(object):
    def __init__(self, value="value"):
        self.value = value
        raise ValueError("Error in constructor")


class WithHelpArg(object):
    """Test class for testing when class has a help= arg."""

    def __init__(self, help=True):  # pylint: disable=redefined-builtin
        self.has_help = help
        self.dictionary = {"__help": "help in a dict"}


class NoDefaults(object):
    def double(self, count):
        return 2 * count

    def triple(self, count):
        return 3 * count


class WithDefaults(object):
    """Class with functions that have default arguments."""

    def double(self, count=0):
        """Returns the input multiplied by 2.

        Args:
          count: Input number that you want to double.

        Returns:
          A number that is the double of count.
        """
        return 2 * count

    def triple(self, count=0):
        return 3 * count

    def text(self, string=("0001020304050607080910111213141516171819" "2021222324252627282930313233343536373839")):
        return string


class OldStyleWithDefaults:  # pylint: disable=old-style-class,no-init
    def double(self, count=0):
        return 2 * count

    def triple(self, count=0):
        return 3 * count


class MixedDefaults(object):
    def ten(self):
        return 10

    def sum(self, alpha=0, beta=0):
        return alpha + 2 * beta

    def identity(self, alpha, beta="0"):
        return alpha, beta


class SimilarArgNames(object):
    def identity(self, bool_one=False, bool_two=False):
        return bool_one, bool_two

    def identity2(self, a=None, alpha=None):
        return a, alpha


class CapitalizedArgNames(object):
    def sum(self, Delta=1.0, Gamma=2.0):  # pylint: disable=invalid-name
        return Delta + Gamma


class Annotations(object):
    def double(self, count=0):
        return 2 * count

    def triple(self, count=0):
        return 3 * count

    double.__annotations__ = {"count": float}
    triple.__annotations__ = {"count": float}


class TypedProperties(object):
    """Test class for testing Python Fire with properties of various types."""

    def __init__(self):
        self.alpha = True
        self.beta = (1, 2, 3)
        self.charlie = WithDefaults()
        self.delta = {
            "echo": "E",
            "nest": {
                0: "a",
                1: "b",
            },
        }
        self.echo = ["alex", "bethany"]
        self.fox = ("carry", "divide")
        self.gamma = "myexcitingstring"


class VarArgs(object):
    """Test class for testing Python Fire with a property with varargs."""

    def cumsums(self, *items):
        total = None
        sums = []
        for item in items:
            if total is None:
                total = item
            else:
                total += item
            sums.append(total)
        return sums

    def varchars(self, alpha=0, beta=0, *chars):  # pylint: disable=keyword-arg-before-vararg
        return alpha, beta, "".join(chars)


class Underscores(object):
    def __init__(self):
        self.underscore_example = "fish fingers"

    def underscore_function(self, underscore_arg):
        return underscore_arg


class BoolConverter(object):
    def as_bool(self, arg=False):
        return bool(arg)


class ReturnsObj(object):
    def get_obj(self, *items):
        del items  # Unused
        return BoolConverter()


class NumberDefaults(object):
    def reciprocal(self, divisor=10.0):
        return 1.0 / divisor

    def integer_reciprocal(self, divisor=10):
        return 1.0 / divisor


class InstanceVars(object):
    def __init__(self, arg1, arg2):
        self.arg1 = arg1
        self.arg2 = arg2

    def run(self, arg1, arg2):
        return (self.arg1, self.arg2, arg1, arg2)


class Kwargs(object):
    def props(self, **kwargs):
        return kwargs

    def upper(self, **kwargs):
        return " ".join(sorted(kwargs.keys())).upper()

    def run(self, positional, named=None, **kwargs):
        return (positional, named, kwargs)


class ErrorRaiser(object):
    def fail(self):
        raise ValueError("This error is part of a test.")


class NonComparable(object):
    def __eq__(self, other):
        raise ValueError("Instances of this class cannot be compared.")

    def __ne__(self, other):
        raise ValueError("Instances of this class cannot be compared.")


class EmptyDictOutput(object):
    def totally_empty(self):
        return {}

    def nothing_printable(self):
        return {"__do_not_print_me": 1}


class CircularReference(object):
    def create(self):
        x = {}
        x["y"] = x
        return x


class OrderedDictionary(object):
    def empty(self):
        return collections.OrderedDict()

    def non_empty(self):
        ordered_dict = collections.OrderedDict()
        ordered_dict["A"] = "A"
        ordered_dict[2] = 2
        return ordered_dict


class NamedTuple(object):
    """Functions returning named tuples used for testing."""

    def point(self):
        """Point example straight from Python docs."""
        # pylint: disable=invalid-name
        Point = collections.namedtuple("Point", ["x", "y"])
        return Point(11, y=22)

    def matching_names(self):
        """Field name equals value."""
        # pylint: disable=invalid-name
        Point = collections.namedtuple("Point", ["x", "y"])
        return Point(x="x", y="y")


class CallableWithPositionalArgs(object):
    """Test class for supporting callable."""

    TEST = 1

    def __call__(self, x, y):
        return x + y

    def fn(self, x):
        return x + 1


NamedTuplePoint = collections.namedtuple("NamedTuplePoint", ["x", "y"])


class SubPoint(NamedTuplePoint):
    """Used for verifying subclasses of namedtuples behave as intended."""

    def coordinate_sum(self):
        return self.x + self.y


class CallableWithKeywordArgument(object):
    """Test class for supporting callable."""

    def __call__(self, **kwargs):
        for key, value in kwargs.items():
            print("%s: %s" % (key, value))

    def print_msg(self, msg):
        print(msg)


class CallableWithTypedKeywordArgument(object):
    """Test class for supporting typed callable."""

    def IntType(self, foo: int):
        self.print_with_type(foo, "foo")

    def StrType(self, foo: str):
        self.print_with_type(foo, "foo")

    def DatetimeType(self, foo: datetime):
        self.print_with_type(foo, "foo")

    def DictType(self, foo: Dict[str, int]):
        self.print_with_type(foo, "foo")

    def DictInDictType(self, foo: Dict[str, Dict[int, datetime]]):
        self.print_with_type(foo, "foo")

    def UnionType(self, foo: Union[int, datetime, str]):
        self.print_with_type(foo, "foo")

    def OptionalType(self, foo: Optional[datetime]):
        self.print_with_type(foo, "foo")

    def DictOptionalUnion(self, foo: Union[Dict[str, Optional[int]], Optional[str]]):
        self.print_with_type(foo, "foo")

    def TupleType(self, foo: Tuple[int, str]):
        self.print_with_type(foo, "foo")

    def TupleOptionalType(self, foo: Tuple[int, Optional[str]]):
        self.print_with_type(foo, "foo")

    def ListType(self, foo: List[str]):
        self.print_with_type(foo, "foo")

    def SetType(self, foo: Set[str]):
        self.print_with_type(foo, "foo")

    def BytesType(self, foo: bytes):
        self.print_with_type(foo, "foo")

    def ListUnionType(self, foo: List[Union[datetime, str]]):
        self.print_with_type(foo, "foo")

    def print_with_type(self, i, *args):
        print(i, type(i).__name__, *args)


CALLABLE_WITH_KEYWORD_ARGUMENT = CallableWithKeywordArgument()


class ClassWithDocstring(object):
    """Test class for testing help text output.

    This is some detail description of this test class.
    """

    def __init__(self, message="Hello!"):
        """Constructor of the test class.

        Constructs a new ClassWithDocstring object.

        Args:
          message: The default message to print.
        """
        self.message = message

    def print_msg(self, msg=None):
        """Prints a message."""
        if msg is None:
            msg = self.message
        print(msg)


class ClassWithMultilineDocstring(object):
    """Test class for testing help text output with multiline docstring.

    This is a test class that has a long docstring description that spans across
    multiple lines for testing line breaking in help text.
    """

    @staticmethod
    def example_generator(n):
        """Generators have a ``Yields`` section instead of a ``Returns`` section.

        Args:
            n (int): The upper limit of the range to generate, from 0 to `n` - 1.

        Yields:
            int: The next number in the range of 0 to `n` - 1.

        Examples:
            Examples should be written in doctest format, and should illustrate how
            to use the function.

            >>> print([i for i in example_generator(4)])
            [0, 1, 2, 3]

        """
        for i in range(n):
            yield i


def simple_set():
    return {1, 2, "three"}


def simple_frozenset():
    return frozenset({1, 2, "three"})


class Subdict(dict):
    """A subclass of dict, for testing purposes."""


# An example subdict.
SUBDICT = Subdict({1: 2, "red": "blue"})


class Color(enum.Enum):
    RED = 1
    GREEN = 2
    BLUE = 3


class HasStaticAndClassMethods(object):
    """A class with a static method and a class method."""

    CLASS_STATE = 1

    def __init__(self, instance_state):
        self.instance_state = instance_state

    @staticmethod
    def static_fn(args):
        return args

    @classmethod
    def class_fn(cls, args):
        return args + cls.CLASS_STATE


def function_with_varargs(arg1, arg2, arg3=1, *varargs):  # pylint: disable=keyword-arg-before-vararg
    """Function with varargs.

    Args:
      arg1: Position arg docstring.
      arg2: Position arg docstring.
      arg3: Flags docstring.
      *varargs: Accepts unlimited positional args.
    Returns:
      The unlimited positional args.
    """
    del arg1, arg2, arg3  # Unused.
    return varargs


def function_with_keyword_arguments(arg1, arg2=3, **kwargs):
    del arg2  # Unused.
    return arg1, kwargs


def fn_with_code_in_docstring():
    """This has code in the docstring.



    Example:
      x = fn_with_code_in_docstring()
      indentation_matters = True



    Returns:
      True.
    """
    return True


class BinaryCanvas(object):
    """A canvas with which to make binary art, one bit at a time."""

    def __init__(self, size=10):
        self.pixels = [[0] * size for _ in range(size)]
        self._size = size
        self._row = 0  # The row of the cursor.
        self._col = 0  # The column of the cursor.

    def __str__(self):
        return "\n".join(" ".join(str(pixel) for pixel in row) for row in self.pixels)

    def show(self):
        print(self)
        return self

    def move(self, row, col):
        self._row = row % self._size
        self._col = col % self._size
        return self

    def on(self):
        return self.set(1)

    def off(self):
        return self.set(0)

    def set(self, value):
        self.pixels[self._row][self._col] = value
        return self


class DefaultMethod(object):
    def double(self, number):
        return 2 * number

    def __getattr__(self, name):
        def _missing():
            return "Undefined function"

        return _missing


class InvalidProperty(object):
    def double(self, number):
        return 2 * number

    @property
    def prop(self):
        raise ValueError("test")


def simple_decorator(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        return f(*args, **kwargs)

    return wrapper


@simple_decorator
def decorated_method(name="World"):
    return "Hello %s" % name


# pylint: disable=g-doc-args,g-doc-return-or-yield
def fn_with_kwarg(arg1, arg2, **kwargs):
    """Function with kwarg.

    :param arg1: Description of arg1.
    :param arg2: Description of arg2.
    :key arg3: Description of arg3.
    """
    del arg1, arg2
    return kwargs.get("arg3")


def fn_with_kwarg_and_defaults(arg1, arg2, opt=True, **kwargs):
    """Function with kwarg and defaults.

    :param arg1: Description of arg1.
    :param arg2: Description of arg2.
    :key arg3: Description of arg3.
    """
    del arg1, arg2, opt
    return kwargs.get("arg3")


# pylint: enable=g-doc-args,g-doc-return-or-yield


def fn_with_multiple_defaults(first="first", last="last", late="late"):
    """Function with kwarg and defaults.

    :key first: Description of first.
    :key last: Description of last.
    :key late: Description of late.
    """
    del last, late
    return first



# Copyright (C) 2018 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the interact module."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from afire import interact
from afire import testutils

import mock


try:
    import IPython  # pylint: disable=unused-import, g-import-not-at-top

    INTERACT_METHOD = "IPython.start_ipython"
except ImportError:
    INTERACT_METHOD = "code.InteractiveConsole"


class InteractTest(testutils.BaseTestCase):
    @mock.patch(INTERACT_METHOD)
    def testInteract(self, mock_interact_method):
        self.assertFalse(mock_interact_method.called)
        interact.Embed({})
        self.assertTrue(mock_interact_method.called)

    @mock.patch(INTERACT_METHOD)
    def testInteractVariables(self, mock_interact_method):
        self.assertFalse(mock_interact_method.called)
        interact.Embed(
            {
                "count": 10,
                "mock": mock,
            }
        )
        self.assertTrue(mock_interact_method.called)


if __name__ == "__main__":
    testutils.main()



# -*- coding: utf-8 -*- #
# Copyright 2018 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Semantic text objects that are used for styled outputting."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import enum


class TextAttributes(object):
    """Attributes to use to style text with."""

    def __init__(self, format_str=None, color=None, attrs=None):
        """Defines a set of attributes for a piece of text.

        Args:
          format_str: (str), string that will be used to format the text
            with. For example '[{}]', to enclose text in brackets.
          color: (Colors), the color the text should be formatted with.
          attrs: (Attrs), the attributes to apply to text.
        """
        self._format_str = format_str
        self._color = color
        self._attrs = attrs or []

    @property
    def format_str(self):
        return self._format_str

    @property
    def color(self):
        return self._color

    @property
    def attrs(self):
        return self._attrs


class TypedText(object):
    """Text with a semantic type that will be used for styling."""

    def __init__(self, texts, text_type=None):
        """String of text and a corresponding type to use to style that text.

        Args:
         texts: (list[str]), list of strs or TypedText objects
           that should be styled using text_type.
         text_type: (TextTypes), the semantic type of the text that
           will be used to style text.
        """
        self.texts = texts
        self.text_type = text_type

    def __len__(self):
        length = 0
        for text in self.texts:
            length += len(text)
        return length

    def __add__(self, other):
        texts = [self, other]
        return TypedText(texts)

    def __radd__(self, other):
        texts = [other, self]
        return TypedText(texts)


class _TextTypes(enum.Enum):
    """Text types base class that defines base functionality."""

    def __call__(self, *args):
        """Returns a TypedText object using this style."""
        return TypedText(list(args), self)


# TODO: Add more types.
class TextTypes(_TextTypes):
    """Defines text types that can be used for styling text."""

    RESOURCE_NAME = 1
    URL = 2
    USER_INPUT = 3
    COMMAND = 4
    INFO = 5
    URI = 6
    OUTPUT = 7
    PT_SUCCESS = 8
    PT_FAILURE = 9



# -*- coding: utf-8 -*- #
# Copyright 2015 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""OS specific console_attr helper functions."""
# This file contains platform specific code which is not currently handled
# by pytype.
# pytype: skip-file

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import os
import sys

from afire.console import encoding


def GetTermSize():
    """Gets the terminal x and y dimensions in characters.

    _GetTermSize*() helper functions taken from:
      http://stackoverflow.com/questions/263890/

    Returns:
      (columns, lines): A tuple containing the terminal x and y dimensions.
    """
    xy = None
    # Believe the first helper that doesn't bail.
    for get_terminal_size in (_GetTermSizePosix, _GetTermSizeWindows, _GetTermSizeEnvironment, _GetTermSizeTput):
        try:
            xy = get_terminal_size()
            if xy:
                break
        except:  # pylint: disable=bare-except
            pass
    return xy or (80, 24)


def _GetTermSizePosix():
    """Returns the Posix terminal x and y dimensions."""
    # pylint: disable=g-import-not-at-top
    import fcntl

    # pylint: disable=g-import-not-at-top
    import struct

    # pylint: disable=g-import-not-at-top
    import termios

    def _GetXY(fd):
        """Returns the terminal (x,y) size for fd.

        Args:
          fd: The terminal file descriptor.

        Returns:
          The terminal (x,y) size for fd or None on error.
        """
        try:
            # This magic incantation converts a struct from ioctl(2) containing two
            # binary shorts to a (rows, columns) int tuple.
            rc = struct.unpack(b"hh", fcntl.ioctl(fd, termios.TIOCGWINSZ, "junk"))
            return (rc[1], rc[0]) if rc else None
        except:  # pylint: disable=bare-except
            return None

    xy = _GetXY(0) or _GetXY(1) or _GetXY(2)
    if not xy:
        fd = None
        try:
            fd = os.open(os.ctermid(), os.O_RDONLY)
            xy = _GetXY(fd)
        except:  # pylint: disable=bare-except
            xy = None
        finally:
            if fd is not None:
                os.close(fd)
    return xy


def _GetTermSizeWindows():
    """Returns the Windows terminal x and y dimensions."""
    # pylint:disable=g-import-not-at-top
    import struct

    # pylint: disable=g-import-not-at-top
    from ctypes import create_string_buffer

    # pylint:disable=g-import-not-at-top
    from ctypes import windll

    # stdin handle is -10
    # stdout handle is -11
    # stderr handle is -12

    h = windll.kernel32.GetStdHandle(-12)
    csbi = create_string_buffer(22)
    if not windll.kernel32.GetConsoleScreenBufferInfo(h, csbi):
        return None
    (
        unused_bufx,
        unused_bufy,
        unused_curx,
        unused_cury,
        unused_wattr,
        left,
        top,
        right,
        bottom,
        unused_maxx,
        unused_maxy,
    ) = struct.unpack(b"hhhhHhhhhhh", csbi.raw)
    x = right - left + 1
    y = bottom - top + 1
    return (x, y)


def _GetTermSizeEnvironment():
    """Returns the terminal x and y dimensions from the environment."""
    return (int(os.environ["COLUMNS"]), int(os.environ["LINES"]))


def _GetTermSizeTput():
    """Returns the terminal x and y dimensions from tput(1)."""
    import subprocess  # pylint: disable=g-import-not-at-top

    output = encoding.Decode(subprocess.check_output(["tput", "cols"], stderr=subprocess.STDOUT))
    cols = int(output)
    output = encoding.Decode(subprocess.check_output(["tput", "lines"], stderr=subprocess.STDOUT))
    rows = int(output)
    return (cols, rows)


_ANSI_CSI = "\x1b"  # ANSI control sequence indicator (ESC)
_CONTROL_D = "\x04"  # unix EOF (^D)
_CONTROL_Z = "\x1a"  # Windows EOF (^Z)
_WINDOWS_CSI_1 = "\x00"  # Windows control sequence indicator #1
_WINDOWS_CSI_2 = "\xe0"  # Windows control sequence indicator #2


def GetRawKeyFunction():
    """Returns a function that reads one keypress from stdin with no echo.

    Returns:
      A function that reads one keypress from stdin with no echo or a function
      that always returns None if stdin does not support it.
    """
    # Believe the first helper that doesn't bail.
    for get_raw_key_function in (_GetRawKeyFunctionPosix, _GetRawKeyFunctionWindows):
        try:
            return get_raw_key_function()
        except:  # pylint: disable=bare-except
            pass
    return lambda: None


def _GetRawKeyFunctionPosix():
    """_GetRawKeyFunction helper using Posix APIs."""
    # pylint: disable=g-import-not-at-top
    import tty

    # pylint: disable=g-import-not-at-top
    import termios

    def _GetRawKeyPosix():
        """Reads and returns one keypress from stdin, no echo, using Posix APIs.

        Returns:
          The key name, None for EOF, <*> for function keys, otherwise a
          character.
        """
        ansi_to_key = {
            "A": "<UP-ARROW>",
            "B": "<DOWN-ARROW>",
            "D": "<LEFT-ARROW>",
            "C": "<RIGHT-ARROW>",
            "5": "<PAGE-UP>",
            "6": "<PAGE-DOWN>",
            "H": "<HOME>",
            "F": "<END>",
            "M": "<DOWN-ARROW>",
            "S": "<PAGE-UP>",
            "T": "<PAGE-DOWN>",
        }

        # Flush pending output. sys.stdin.read() would do this, but it's explicitly
        # bypassed in _GetKeyChar().
        sys.stdout.flush()

        fd = sys.stdin.fileno()

        def _GetKeyChar():
            return encoding.Decode(os.read(fd, 1))

        old_settings = termios.tcgetattr(fd)
        try:
            tty.setraw(fd)
            c = _GetKeyChar()
            if c == _ANSI_CSI:
                c = _GetKeyChar()
                while True:
                    if c == _ANSI_CSI:
                        return c
                    if c.isalpha():
                        break
                    prev_c = c
                    c = _GetKeyChar()
                    if c == "~":
                        c = prev_c
                        break
                return ansi_to_key.get(c, "")
        except:  # pylint:disable=bare-except
            c = None
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
        return None if c in (_CONTROL_D, _CONTROL_Z) else c

    return _GetRawKeyPosix


def _GetRawKeyFunctionWindows():
    """_GetRawKeyFunction helper using Windows APIs."""
    # pylint: disable=g-import-not-at-top
    import msvcrt

    def _GetRawKeyWindows():
        """Reads and returns one keypress from stdin, no echo, using Windows APIs.

        Returns:
          The key name, None for EOF, <*> for function keys, otherwise a
          character.
        """
        windows_to_key = {
            "H": "<UP-ARROW>",
            "P": "<DOWN-ARROW>",
            "K": "<LEFT-ARROW>",
            "M": "<RIGHT-ARROW>",
            "I": "<PAGE-UP>",
            "Q": "<PAGE-DOWN>",
            "G": "<HOME>",
            "O": "<END>",
        }

        # Flush pending output. sys.stdin.read() would do this it's explicitly
        # bypassed in _GetKeyChar().
        sys.stdout.flush()

        def _GetKeyChar():
            return encoding.Decode(msvcrt.getch())

        c = _GetKeyChar()
        # Special function key is a two character sequence; return the second char.
        if c in (_WINDOWS_CSI_1, _WINDOWS_CSI_2):
            return windows_to_key.get(_GetKeyChar(), "")
        return None if c in (_CONTROL_D, _CONTROL_Z) else c

    return _GetRawKeyWindows



# Copyright (C) 2018 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test the test utilities for Fire's tests."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys

from afire import testutils

import six


class TestTestUtils(testutils.BaseTestCase):
    """Let's get meta."""

    def testNoCheckOnException(self):
        with self.assertRaises(ValueError):
            with self.assertOutputMatches(stdout="blah"):
                raise ValueError()

    def testCheckStdoutOrStderrNone(self):
        with six.assertRaisesRegex(self, AssertionError, "stdout:"):
            with self.assertOutputMatches(stdout=None):
                print("blah")

        with six.assertRaisesRegex(self, AssertionError, "stderr:"):
            with self.assertOutputMatches(stderr=None):
                print("blah", file=sys.stderr)

        with six.assertRaisesRegex(self, AssertionError, "stderr:"):
            with self.assertOutputMatches(stdout="apple", stderr=None):
                print("apple")
                print("blah", file=sys.stderr)

    def testCorrectOrderingOfAssertRaises(self):
        # Check to make sure FireExit tests are correct.
        with self.assertOutputMatches(stdout="Yep.*first.*second"):
            with self.assertRaises(ValueError):
                print("Yep, this is the first line.\nThis is the second.")
                raise ValueError()


if __name__ == "__main__":
    testutils.main()



# Copyright (C) 2018 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the trace module."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from afire import testutils
from afire import trace


class FireTraceTest(testutils.BaseTestCase):
    def testFireTraceInitialization(self):
        t = trace.FireTrace(10)
        self.assertIsNotNone(t)
        self.assertIsNotNone(t.elements)

    def testFireTraceGetResult(self):
        t = trace.FireTrace("start")
        self.assertEqual(t.GetResult(), "start")
        t.AddAccessedProperty("t", "final", None, "example.py", 10)
        self.assertEqual(t.GetResult(), "t")

    def testFireTraceHasError(self):
        t = trace.FireTrace("start")
        self.assertFalse(t.HasError())
        t.AddAccessedProperty("t", "final", None, "example.py", 10)
        self.assertFalse(t.HasError())
        t.AddError(ValueError("example error"), ["arg"])
        self.assertTrue(t.HasError())

    def testAddAccessedProperty(self):
        t = trace.FireTrace("initial object")
        args = ("example", "args")
        t.AddAccessedProperty("new component", "prop", args, "sample.py", 12)
        self.assertEqual(str(t), '1. Initial component\n2. Accessed property "prop" (sample.py:12)')

    def testAddCalledCallable(self):
        t = trace.FireTrace("initial object")
        args = ("example", "args")
        t.AddCalledComponent("result", "cell", args, "sample.py", 10, False, action=trace.CALLED_CALLABLE)
        self.assertEqual(str(t), '1. Initial component\n2. Called callable "cell" (sample.py:10)')

    def testAddCalledRoutine(self):
        t = trace.FireTrace("initial object")
        args = ("example", "args")
        t.AddCalledComponent("result", "run", args, "sample.py", 12, False, action=trace.CALLED_ROUTINE)
        self.assertEqual(str(t), '1. Initial component\n2. Called routine "run" (sample.py:12)')

    def testAddInstantiatedClass(self):
        t = trace.FireTrace("initial object")
        args = ("example", "args")
        t.AddCalledComponent("Classname", "classname", args, "sample.py", 12, False, action=trace.INSTANTIATED_CLASS)
        target = """1. Initial component
2. Instantiated class "classname" (sample.py:12)"""
        self.assertEqual(str(t), target)

    def testAddCompletionScript(self):
        t = trace.FireTrace("initial object")
        t.AddCompletionScript("This is the completion script string.")
        self.assertEqual(str(t), "1. Initial component\n2. Generated completion script")

    def testAddInteractiveMode(self):
        t = trace.FireTrace("initial object")
        t.AddInteractiveMode()
        self.assertEqual(str(t), "1. Initial component\n2. Entered interactive mode")

    def testGetCommand(self):
        t = trace.FireTrace("initial object")
        args = ("example", "args")
        t.AddCalledComponent("result", "run", args, "sample.py", 12, False, action=trace.CALLED_ROUTINE)
        self.assertEqual(t.GetCommand(), "example args")

    def testGetCommandWithQuotes(self):
        t = trace.FireTrace("initial object")
        args = ("example", "spaced arg")
        t.AddCalledComponent("result", "run", args, "sample.py", 12, False, action=trace.CALLED_ROUTINE)
        self.assertEqual(t.GetCommand(), "example 'spaced arg'")

    def testGetCommandWithFlagQuotes(self):
        t = trace.FireTrace("initial object")
        args = ("--example=spaced arg",)
        t.AddCalledComponent("result", "run", args, "sample.py", 12, False, action=trace.CALLED_ROUTINE)
        self.assertEqual(t.GetCommand(), "--example='spaced arg'")


class FireTraceElementTest(testutils.BaseTestCase):
    def testFireTraceElementHasError(self):
        el = trace.FireTraceElement()
        self.assertFalse(el.HasError())

        el = trace.FireTraceElement(error=ValueError("example error"))
        self.assertTrue(el.HasError())

    def testFireTraceElementAsStringNoMetadata(self):
        el = trace.FireTraceElement(
            component="Example",
            action="Fake action",
        )
        self.assertEqual(str(el), "Fake action")

    def testFireTraceElementAsStringWithTarget(self):
        el = trace.FireTraceElement(
            component="Example",
            action="Created toy",
            target="Beaker",
        )
        self.assertEqual(str(el), 'Created toy "Beaker"')

    def testFireTraceElementAsStringWithTargetAndLineNo(self):
        el = trace.FireTraceElement(
            component="Example",
            action="Created toy",
            target="Beaker",
            filename="beaker.py",
            lineno=10,
        )
        self.assertEqual(str(el), 'Created toy "Beaker" (beaker.py:10)')


if __name__ == "__main__":
    testutils.main()
